"""
A collection of utility functions to help verify the integrity of collected data and clean the checkpoints/metrics.
"""
import logging
import shutil, sys
import pickle
from functools import partial
from pathlib import Path
from typing import Any, Optional, Generator, Tuple, Union, List, Dict

import pandas as pd
import torch

from jahs_bench.tabular.lib.core.utils import DirectoryTree, Checkpointer, MetricLogger

_log = logging.getLogger(__name__)

MODIFIED_FLAG_FILENAME = "verification_needed"

def _verify_chkpt(pth: Path, map_location: Optional[Any] = None):
    """ Attempt to load a single checkpoint file specified by the given path. Returns True if the file can be read
    successfully. 'map_location' has been provided for compatibility with GPUs. """

    try:
        _ = torch.load(pth, map_location=map_location)
    except (RuntimeError, EOFError) as e:
        _log.debug(f"Found corrupt checkpoint: {pth}\nError type: {type(e)}\nError description: {str(e)}")
        return False
    except Exception as e:
        _log.info(f"Failed to read checkpoint {pth} due to unhandled error of type {type(e)}, error description: "
                  f"{str(e)}")
        return True
    else:
        return True


def _verify_model_chkpts(dtree: DirectoryTree, cleanup: Optional[bool] = False, backuptree: DirectoryTree = None,
                         map_location: Optional[Any] = None):
    """ Given a DirectoryTree which has been initialized with a particular taskid and model_idx, verifies the
    integrity of all checkpoints under this model. 'map_location' has been provided for compatibility with GPUs. """

    assert dtree.taskid is not None, "The directory tree has not been initialzed with a taskid."
    assert dtree.model_idx is not None, "The directory tree has not been initialzed with a model index."
    assert cleanup and (backuptree is not None), "If cleanup has been enabled, a backup directory tree must be given."

    chkpt_pths = Checkpointer._get_sorted_chkpt_paths(dtree.model_checkpoints_dir)
    for pth in chkpt_pths:
        check = _verify_chkpt(pth)
        if not check and cleanup:
            shutil.move(pth, backuptree.model_checkpoints_dir / pth.name)
            (dtree.model_dir / MODIFIED_FLAG_FILENAME).touch()


def _verify_metrics_log(pth: Path):
    """ Attempt to load a single metrics DataFrame file specified by the given path. Returns True if the file can be
    read successfully. """

    try:
        _ = pd.read_pickle(pth)
    except (pickle.PickleError, EOFError) as e:
        _log.debug(f"Found corrupt metrics log: {pth}\nError type: {type(e)}\nError description: {str(e)}")
        return False
    except Exception as e:
        _log.info(f"Failed to read metrics log {pth} due to unhandled error of type {type(e)}, error description: "
                  f"{str(e)}")
        return True
    else:
        return True


def _verify_model_metrics(dtree: DirectoryTree, cleanup: Optional[bool] = False, backuptree: DirectoryTree = None):
    """ Given a DirectoryTree which has been initialized with a particular taskid and model_idx, verifies the
    integrity of all metric DataFrames under this model. """

    assert dtree.taskid is not None, "The directory tree has not been initialzed with a taskid."
    assert dtree.model_idx is not None, "The directory tree has not been initialzed with a model index."
    assert cleanup and (backuptree is not None), "If cleanup has been enabled, a backup directory tree must be given."

    metric_pths = MetricLogger._get_sorted_metric_paths(dtree.model_metrics_dir)
    for pth in metric_pths:
        check = _verify_metrics_log(pth)
        if not check and cleanup:
            shutil.move(pth, backuptree.model_checkpoints_dir / pth.name)
            (dtree.model_dir / MODIFIED_FLAG_FILENAME).touch()


# TODO: Update return type and documentation
def iterate_model_tree(basedir: Path, taskid: Optional[int] = None, model_idx: Optional[int] = None,
                       enumerate: Optional[bool] = False) -> \
        Union[Generator[DirectoryTree, None, None], Generator[Tuple[int, int, DirectoryTree], None, None]]:
    """
    Iterates over a directory-tree model-wise. If only 'basedir' is given, iterates over every model of every task
    found at this base directory. If a taskid is also specified, iterates over every model of that specific task. If a
    model index is specified in addition to a taskid, iterates over only that single model's directory tree, i.e.
    yields a single directory tree. It is possible to iterate over every model corresponding to a specific model idx in
    every task found at 'basedir' by setting taskid to None and specifying a value for model_idx. Setting 'enumerate'
    to True causes the generator to also yield the current taskid and model_idx.
    """

    # Opening in read-only mode prevents accidentally creating directories that did not already exist.
    dtree = DirectoryTree(basedir=basedir, read_only=True)

    if taskid is not None:
        dtree.taskid = taskid
        task_dirs = [dtree.task_dir]
    else:
        task_dirs = dtree.existing_tasks
        if task_dirs is None:
            _log.info(f"No tasks found at {basedir}.")
            yield StopIteration

    for t in task_dirs:
        dtree.taskid = int(t.stem)

        if model_idx is not None:
            dtree.model_idx = model_idx
            model_dirs = [dtree.model_dir]
        else:
            model_dirs = dtree.existing_models
            if model_dirs is None:
                _log.info(f"No models found in task {dtree.taskid} at {basedir}.")
                continue

        for m in model_dirs:
            dtree.model_idx = int(m.stem)
            # new_tree = DirectoryTree(basedir=dtree.basedir, taskid=dtree.taskid, model_idx=dtree.model_idx,
            #                          read_only=dtree.read_only)
            yield (t, m, dtree) if enumerate else dtree
            dtree.model_idx = None

        dtree.taskid = None


def clean_corrupt_files(basedir: Path, taskid: Optional[int] = None, model_idx: Optional[int] = None,
                        cleanup: Optional[bool] = False, backupdir: Optional[Path] = None,
                        map_location: Optional[Any] = None):
    """
    Attempts to read each checkpoint and metric dataframe file and deletes any that cannot be read. If only 'basedir'
    is given, all tasks and their models found at this base directory will be cleaned. If a taskid is also specified,
    only the models of that specific task are cleaned. If a model index is specified in addition to a taskid, only that
    model's files are cleaned. The flag 'cleanup' has been intentionally introduced as an extra user-level check before
    deleting files - this flag must be set to True to delete corrupted files, otherwise, only debug messages for
    corrupted files are generated. This also enables a "debug mode" usage of this function call, with 'cleanup' set to
    False. 'map_location' has been provided for compatibility with GPUs.
    """

    if cleanup:
        assert backupdir is not None, "When cleanup is enabled, a backup directory must be given."
        backupdir.mkdir(exist_ok=True, parents=True)
        backuptree = DirectoryTree(basedir=backupdir, read_only=True)

    for dtree in iterate_model_tree(basedir=basedir, taskid=taskid, model_idx=model_idx, enumerate=False):
        if cleanup:
            backuptree.taskid = dtree.taskid
            backuptree.model_idx = dtree.model_idx

        _verify_model_chkpts(dtree, cleanup=cleanup, backuptree=backuptree, map_location=map_location)
        _verify_model_metrics(dtree, cleanup=cleanup, backuptree=backuptree)


class MetricDataIntegrityChecker:
    dtree: DirectoryTree

    # found_faults = False  # If faults have been found, there is a need to run a more intense data verification procedure
    # consistent = True  # Should the data be inconsistent, there is no choice but to abandon it
    # faulty_epochs = []  # The epoch nums found to be the cause of data faults
    # verified_chkpts = []  # The checkpoint timestamps that have been verified to be consistent and fault-free
    # verified_metric_logs = []  # The metrics log timestamps that have been verified to be consistent and fault-free
    # fixed_metric_logs = []  # The metrics log timestamps that needed to be fixed
    # invalid_chkpts = []  # The checkpoint timestamps that have been deemed inconsistent and should be deleted
    # invalid_metric_logs = []  # The metrics log timestamps that have been deemed inconsistent and should be deleted

    _map_location = None
    __CHECKPOINT = Dict

    def __init__(self, dtree: DirectoryTree, enable_cleanup: bool = False):
        assert dtree.taskid is not None and dtree.model_idx is not None, \
            "The directory tree must be initialized fully with a task id and model index before being passed. The " \
            f"given tree had task id {dtree.taskid} and model index {dtree.model_idx}."
        self.dtree = dtree
        self.enable_cleanup = enable_cleanup

    # Convenience wrappers
    _load_chkpt = partial(Checkpointer._load_checkpoint, safe_load=True, map_location=_map_location)
    _load_metric_log = partial(MetricLogger._load_metrics_log, safe_load=True)

    @property
    def chkpt_pths(self) -> Dict[float, Path]:
        """ Dictionary mapping float timestamp to checkpoint file Path, sorted in descending order, i.e. most recent
        checkpoint first. """
        return {Checkpointer._extract_runtime_from_filename(p): p for p in
                Checkpointer._get_sorted_chkpt_paths(pth=self.dtree.model_checkpoints_dir, ascending=False)}

    @property
    def chkpt_timestamps(self) -> List[float]:
        """ The list of timestamps for this model's available checkpoints, in chronologically descending order. """
        return list(map(Checkpointer._extract_runtime_from_filename,
                        Checkpointer._get_sorted_chkpt_paths(pth=self.dtree.model_checkpoints_dir, ascending=False)))

    @property
    def metrics_log_pths(self) -> Dict[float, Path]:
        """ Dictionary mapping float timestamp to metrics log file Path, sorted in descending order, i.e. most recent
        log first. """
        return {MetricLogger._extract_runtime_from_filename(p): p for p in
                MetricLogger._get_sorted_metric_paths(pth=self.dtree.model_metrics_dir, ascending=False)}

    @property
    def metrics_log_timestamps(self) -> List[float]:
        """ The list of timestamps for this model's available metrics logs, in chronologically descending order. """
        return list(map(MetricLogger._extract_runtime_from_filename,
                        MetricLogger._get_sorted_metric_paths(pth=self.dtree.model_metrics_dir, ascending=False)))

    # def save_results(self):
    #     """ Saves a JSON file in the model directory detailing the results of the verification procedure. """
    #     save_dict = {
    #         "found_faults": self.found_faults,
    #         "consistent": self.consistent,
    #         "faulty_epochs": self.faulty_epochs,
    #         "verified_chkpts": self.verified_chkpts,
    #         "verified_metric_logs": self.verified_metric_logs,
    #         "fixed_metric_logs": self.fixed_metric_logs,
    #         "invalid_chkpts": self.invalid_chkpts,
    #         "invalid_metric_logs": self.invalid_metric_logs,
    #     }
    #     with open(self.dtree.model_metrics_dir / "metric_logs_verification_results.json") as fp:
    #         json.dump(save_dict, fp)

    # def _declare_invalid(self):
    #     """ Declares all of the current model's data to be invalid and saves the results to disk. """
    #
    #     self.consistent = False
    #     self.invalid_chkpts = list(sorted(self.chkpt_pths.keys(), reverse=True))
    #     self.invalid_metric_logs = list(sorted(self.metrics_log_pths.keys(), reverse=True))
    #     save_results()

    def _iterate_corresponding_pairs_over_timestamps(
            self, start: Optional[int] = 0, stop: Optional[int] = None, step: Optional[int] = 1,
            timestamps: bool = False, nepochs: bool = False) -> \
            Generator[Union[Tuple[__CHECKPOINT, pd.DataFrame],
                            Tuple[__CHECKPOINT, pd.DataFrame, float],
                            Tuple[__CHECKPOINT, pd.DataFrame, int, int],
                            Tuple[__CHECKPOINT, pd.DataFrame, float, int, int]], None, None]:
        """ Iterates over all the known timestamps in chronologically descending order and yields the
        corresponding checkpoint and the metrics log. If 'timestamps' is True, also returns the corresponding timestamp
        as a float. If 'nepochs' is True, also returns the #epochs in checkpoint and #epochs in metrics log. For the
        cases when either the checkpoint or the metric log is missing, None is returned for that value. Only the
        timestamps of the checkpoints are used. """

        known_timestamps = list(sorted(set(self.chkpt_timestamps + self.metrics_log_timestamps), reverse=True))
        chkpt_pths = self.chkpt_pths
        metrics_log_pths = self.metrics_log_pths

        for timestamp in known_timestamps[start:stop:step]:
            if timestamp in chkpt_pths:
                chkpt = self._load_chkpt(chkpt_pths[timestamp])
                nepochs_chkpt = chkpt["epochs"] + 1
            else:
                chkpt = None
                nepochs_chkpt = None

            if timestamp in metrics_log_pths:
                metrics_log = self._load_metric_log(metrics_log_pths[timestamp])
                nepochs_metrics = metrics_log.index.size
            else:
                metrics_log = None
                nepochs_metrics = None

            if timestamps and nepochs:
                yield chkpt, metrics_log, timestamp, nepochs_chkpt, nepochs_metrics
            elif timestamps:
                yield chkpt, metrics_log, timestamp
            elif nepochs:
                yield chkpt, metrics_log, nepochs_chkpt, nepochs_metrics
            else:
                yield chkpt, metrics_log

            del chkpt
            del metrics_log

    # def _is_all_metric_data_clean(self) -> bool:
    #     """ Checks if, for corresponding checkpoint and metric log pairs only, all existing pairs agree on the number
    #     of registered epochs. """
    #
    #     gen = self._iterate_corresponding_pairs_over_timestamps(timestamps=False, nepochs=True)
    #     for chkpt, metrics_log, nepochs_chkpt, nepochs_metrics in gen:
    #         if chkpt is None or metrics_log is None:
    #             continue
    #         if nepochs_chkpt != nepochs_metrics:
    #             _log.debug(f"Checkpoint {timestamp} of model at {self.dtree.model_dir} is not consistent in terms of "
    #                        f"number of epochs with its corresponding metric logs.")
    #             del gen
    #             return False
    #     return True

    # @classmethod
    # def _fix_metric_logs_faulty_epochs(cls, metrics_log: pd.DataFrame, faulty_epochs: Iterable[int]) -> pd.DataFrame:
    #     """ Removes the rows corresponding to indices listed in 'faulty_epochs' from 'metrics_log', re-indexes the
    #     DataFrame and returns it. """
    #
    #     metrics_log: pd.DataFrame = metrics_log.drop(index=faulty_epochs)
    #     nepochs = metrics_log.index.size
    #     metrics_log.index = pd.Int64Index(list(range(nepochs))) + 1
    #     return metrics_log

    def check_model_metric_data_integrity_aggressive(self, backup_dir: Path):
        """
        Performs a more aggressive data integrity check for a single model config..
        How it works:
        1 - For each checkpoint, load the corresponding metric log, starting from the most recent. This is guaranteed
            to exist only in the case when the checkpoint registers end of training epochs have been reached and is the
            bug that has prompted this script to be written.
        2 - Iterate through the checkpoint timestamps in chronologically descending order until the most recent such
            timestamp is found, wherein both a metric log and a checkpoint for that timestamp exist and they agree on
            the number of epochs for which data has been collected. Save this pair and continue searching.
        3 - If all other pairs agree on the number of epochs, we can be certain that only the data beyond this point
            has been corrupted. Move all the more recent checkpoints as well as any stray metric logs to the backup
            directory. The backup directory serves as a base directory for a duplicate directory tree.
        4 - Mark the existing checkpoints for stage 2 verification.

        During stage 2 verification, each checkpoint is loaded in chronologically ascending order and trained for
        1-2 epochs. Then, the next most recent metric log is loaded and the results of these epochs is cross-checked.
        If the results don't add up, all subsequent benchmarks and metric logs are also removed. If there is no metric
        log for the current benchmark, it, too, is discarded. Thus, only verifiably consistent data is left.
        """

        all_timestamps = []
        stray_metric_logs = []
        dangling_chkpts = {}
        existing_pairs = []
        consistent_pairs = []
        inconsistent_pairs = []

        gen = self._iterate_corresponding_pairs_over_timestamps(timestamps=True, nepochs=True)
        for chkpt, metrics_log, timestamp, nepochs_chkpt, nepochs_metrics in gen:
            all_timestamps.append(timestamp)
            if chkpt is None and metrics_log is not None:
                stray_metric_logs.append(timestamp)
            elif chkpt is not None and metrics_log is None:
                dangling_chkpts[timestamp] = nepochs_chkpt
            elif chkpt is not None and metrics_log is not None:
                existing_pairs.append(timestamp)

                if nepochs_chkpt == nepochs_metrics:
                    consistent_pairs.append(timestamp)
                else:
                    inconsistent_pairs.append(timestamp)

        if len(all_timestamps) == 0:
            _log.debug(f"No timestamps to verify in {self.dtree.model_dir}.")
            return

        _log.debug(f"Discovered {len(all_timestamps)} timestamps to verify in {self.dtree.model_dir}.")
        if inconsistent_pairs:
            first_inconsistent_timestamp = min(inconsistent_pairs)
            _log.debug(f"Discovered the first inconsistent pair at timestamp {first_inconsistent_timestamp}")
        else:
            first_inconsistent_timestamp = max(all_timestamps) + 1.
            _log.debug(f"Discovered no inconsistent pairs of checkpoints and metrics logs.")

        # All data after first_inconst is corrupt, including any corresponding pairs
        # All strays should be deleted
        # For any remaining dangling checkpoints, create inferred metric logs and verify them
        # Perform checkpoint verification on all checkpoints?

        to_be_deleted = []
        if first_inconsistent_timestamp != -1.:
            to_be_deleted = [t for t in all_timestamps if t >= first_inconsistent_timestamp]
        to_be_deleted = set(to_be_deleted + stray_metric_logs)
        to_be_inferred = set(dangling_chkpts).difference(to_be_deleted)
        existing_pairs = set(existing_pairs).difference(to_be_deleted)

        chkpt_pths = self.chkpt_pths
        metrics_log_pths = self.metrics_log_pths

        modified = False

        _log.debug(f"Attempting to infer {len(to_be_inferred)} metric logs.")
        inferred = 0
        for t in to_be_inferred:
            closest = [pairt for pairt in existing_pairs if t < pairt]
            if closest:
                closest = min(closest)
            else:
                _log.debug(f"No metric logs available to infer metrics for checkpoint at {t}")
                to_be_deleted.add(t)
                continue
            required_epochs = dangling_chkpts[t]
            closest_metrics_log = self._load_metric_log(metrics_log_pths[closest])
            new_log = closest_metrics_log.iloc[:required_epochs]
            new_log.to_pickle(self.dtree.model_metrics_dir / f"{t}.pkl.gz")
            modified = True
            inferred += 1

        _log.info(f"Inferred {inferred} metrics logs from model {self.dtree.model_dir}.")

        backup_tree = DirectoryTree(basedir=backup_dir, taskid=self.dtree.taskid, model_idx=self.dtree.model_idx,
                                    read_only=True)
        _log.debug(f"Moving {len(to_be_deleted)} checkpoints/metrics logs to the backup directory "
                   f"{backup_tree.model_dir}.")
        deleted = 0
        for t in to_be_deleted:
            if t in chkpt_pths:
                shutil.move(src=chkpt_pths[t], dst=backup_tree.model_checkpoints_dir / chkpt_pths[t].name)
            if t in metrics_log_pths:
                shutil.move(src=metrics_log_pths[t], dst=backup_tree.model_metrics_dir / metrics_log_pths[t].name)
            modified = True
            deleted += 0

        _log.info(f"Deleted {deleted} invalid metrics logs/checkpoints from model {self.dtree.model_dir}.")

        if modified:
            (self.dtree.model_dir / MODIFIED_FLAG_FILENAME).touch()


        # TODO: Add stage 2 verification

    # def _locate_faults(self, first_inconsistent_timestamp: float, last_consistent_timestamp: float,
    #                    n_faulty_epochs: int):
    #     """ Attempts to locate a stray metrics log file that may have introduced inconsistencies in the logged data,
    #     and retroactively removes the faulty data. """
    #
    #     chkpt_pths = self.chkpt_pths
    #     metrics_log_pths = self.metrics_log_pths
    #
    #     timestamps = self.metrics_log_timestamps
    #     idx_first_inconst = timestamps.index(first_inconsistent_timestamp)
    #     idx_last_const = timestamps.index(last_consistent_timestamp)
    #
    #     assert idx_last_const > idx_first_inconst, \
    #         "The first inconsistent pair cannot occur before the last consistent pair."
    #
    #     source_inconsistency = timestamps[idx_first_inconst:idx_last_const][0]
    #     assert source_inconsistency <= first_inconsistent_timestamp, \
    #         "Expected the timestamps to be arranged in descending order."
    #
    #     problematic_metrics_log = self._load_metric_log(metrics_log_pths[source_inconsistency])
    #     nepochs_source_inconst = problematic_metrics_log.index.size
    #     del problematic_metrics_log
    #
    #     last_consistent_chkpt = self._load_chkpt(chkpt_pths[last_consistent_timestamp])
    #     nepochs_last_consistent = last_consistent_chkpt["epochs"]
    #     del last_consistent_chkpt
    #
    #     if n_faulty_epochs != nepochs_source_inconst - nepochs_last_consistent:
    #         _log.info(f"Found irrecoverably inconsistent data at {self.dtree.model_dir}.")
    #         self._declare_invalid()
    #     else:
    #         # Try to fix inconsistencies in all checkpoints moving from this point onwards
    #         timestamps = self.chkpt_timestamps
    #         faulty_epochs = list(range(nepochs_last_consistent + 1, nepochs_last_consistent + n_faulty_epochs + 1))
    #
    #         for timestamp in timestamps[idx_first_inconst:]:
    #             metrics_log = self._load_metric_log(metrics_log_pths[timestamp])
    #             metrics_log = self._fix_metric_logs_faulty_epochs(metrics_log, faulty_epochs=faulty_epochs)
    #             metrics_log.to_pickle(metrics_log_pths[timestamp])
    #
    #         self.faulty_epochs += faulty_epochs
    #
    #     return
    #
    # def check_model_metric_data_integrity(self):
    #     """
    #     Performs the actual data integrity check for a single model config on behalf of
    #     check_metric_data_integrity().
    #     How it works:
    #     1 - For each checkpoint, load the corresponding metric log, starting from the most recent.
    #     2 - If no inconsistency is found, move to (7).
    #     3 - If an inconsistency in the number of epochs is found, keep loading older checkpoints until they become
    #         consistent again. Mark this timestamp as t_const and the oldest inconsistent one as t_inconst.
    #     4 - Assume that the most recent metric log older than t_inconst is the one that caused the inconsistency.
    #         Therefore, deleting the corresponding epochs should remove all inconsistency.
    #     5 - Act on this assumption to try and fix all known inconsistent metric logs.
    #     6 - Start from 1 again.
    #     7 - If any inconsistencies were found, mark this model for cross-verification and delete all metric logs
    #         without corresponding checkpoint files.
    #     """
    #
    #     ## Part 1, look for faulty data
    #     timestamps = self.chkpt_timestamps
    #     first_inconsistent_timestamp = max(timestamps) + 1.
    #     last_consistent_timestamp = -1
    #     n_faulty_epochs = 0
    #
    #     # Iterate over all checkpoint/metrics log pairs by checkpoint timestamp and locate the more recent pair which
    #     # is consistent and the oldest pair which is not consistent. Consistency is defined as having a matching number
    #     # of registered epochs.
    #     gen = self._iterate_corresponding_pairs_over_timestamps(timestamps=True, nepochs=True)
    #     for chkpt, metrics_log, timestamp, nepochs_chkpt, nepochs_metrics in gen:
    #         if nepochs_chkpt == nepochs_metrics:
    #             last_consistent_timestamp = max(timestamp, last_consistent_timestamp)
    #         elif nepochs_chkpt > nepochs_metrics:
    #             raise RuntimeError(f"Unexpected inconsistency type encountered in {self.dtree.model_dir} at timestamp "
    #                                f"{timestamp}.")
    #         else:
    #             first_inconsistent_timestamp = min(timestamp, first_inconsistent_timestamp)
    #             n_faulty_epochs = nepochs_metrics - nepochs_chkpt
    #
    #     if first_inconsistent_timestamp == -1.:
    #         # No inconsistent pairs were found. This model's data is currently clean.
    #         _log.debug(f"Data at {self.dtree.model_dir} verified to be consistent. Moving on to cleanup.")
    #         self._cleanup()
    #         return
    #
    #     elif last_consistent_timestamp == -1.:
    #         # There is no clean data since we checked all available checkpoints and found no consistent pairs.
    #         _log.debug(f"Found irrecoverably inconsistent data at {self.dtree.model_dir}.")
    #         self._declare_invalid()
    #         return
    #
    #     elif last_consistent_timestamp < first_inconsistent_timestamp:
    #         # This is probably a data fault that can be recovered from. Most likely, a checkpoint file was corrupted
    #         # and could not be read, so the training procedure loaded an older checkpoint but a newer metrics log and
    #         # appended the new metrics to the latter.
    #         self.found_faults = True
    #         self._locate_faults(first_inconsistent_timestamp, last_consistent_timestamp, n_faulty_epochs)
    #
    #         # Do another pass of this procedure. This verifies if there are any more data faults and eventually cleans
    #         # up any stray files before exiting.
    #         self.check_model_metric_data_integrity()
    #     else:
    #         # This is some completely unexpected type of data corruption and we have no idea how to solve it.
    #         _log.debug(f"Found irrecoverably inconsistent data at {self.dtree.model_dir}.")
    #         self._declare_invalid()
    #         return
    #
    # def _cleanup(self):
    #     """ If the cleanup flag is set, removes all metrics logs for which a corresponding checkpoint does not exist.
    #     """
    #
    #     if not self.enable_cleanup:
    #         return
    #
    #     all_metrics = self.metrics_log_pths
    #     all_chkpts = self.chkpt_pths
    #
    #     for k, v in all_metrics.items():
    #         if k not in all_chkpts:
    #             _log.debug(f"Removing stray metrics log file {v}")
    #             os.remove(v)


def check_metric_data_integrity(backup_dir: Path, basedir: Path, taskid: Optional[int] = None,
                                model_idx: Optional[int] = None, cleanup: Optional[bool] = False,
                                map_location: Optional[Any] = None):
    """
    This utility checks each logged metric DataFrame's registered number of epochs against those of the checkpoint
    closest to its own timestamp. In most cases, the checkpoint will have an identical timestamp and both the
    checkpoint and the metrics log will have an identical number of epochs. In the cases when an older checkpoint than
    the metrics log must be loaded, it is likely that metric logs recorded after this one contain "false" data -
    data for epochs for which the checkpoint has been lost. In such cases, the probable epoch IDs will be recorded.
    When the next metrics log and corresponding checkpoint are loaded, it is checked if the number of false epochs
    remains consistent or not. In case this consistency is maintained, it is highly likely that all data can be fixed
    by removing the false epochs, so the metric logs are modified and saved accordingly. If inconsistent false epochs
    are found, all checkpoints and metric logs after the last known timestamp at which the checkpoints and metric logs
    were correlated are marked as corrupt.
    """

    backup_dir.mkdir(exist_ok=True, parents=True)
    MetricDataIntegrityChecker._map_location = map_location
    for t, m, dtree in iterate_model_tree(basedir=basedir, taskid=taskid, model_idx=model_idx, enumerate=True):
        checker = MetricDataIntegrityChecker(dtree, enable_cleanup=cleanup)
        # checker.check_model_metric_data_integrity()
        checker.check_model_metric_data_integrity_aggressive(backup_dir=backup_dir)

if __name__ == "__main__":
    ## TEST SCRIPT
    # Setup this module's logger
    fmt = logging.Formatter("[%(asctime)s] %(name)s %(levelname)s: %(message)s", datefmt="%m/%d %H:%M:%S")
    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(fmt)
    _log.addHandler(ch)
    _log.setLevel(logging.DEBUG)

    test_pth = Path("/home/archit/thesis/experiments/test")
    tree_kwargs = {"basedir": test_pth, "taskid": 2, "model_idx": 2}
    # tree_kwargs = {"basedir": test_pth / "N-1-W-4-Resolution-0.25/tasks", "taskid": 0, "model_idx": 1}

    dtree = DirectoryTree(**tree_kwargs)
    # corrupt = True
    corrupt = False
    original_chkpts = Checkpointer._get_sorted_chkpt_paths(dtree.model_checkpoints_dir, ascending=True)
    original_metric_logs = MetricLogger._get_sorted_metric_paths(dtree.model_metrics_dir, ascending=True)


    def introduce_fault(i: int, n: int):
        """ Introduce intentional and controlled data faults in the metric logs. Introduces a new metric log in between
        indices i and i+1, copying over the last 'n' epochs of data from the log at index i into this log. Then,
        re-writes all subsequent logs to mimic this data fault. """

        base_log = MetricLogger._load_metrics_log(original_metric_logs[i])
        new_timestamp = sum([MetricLogger._extract_runtime_from_filename(original_metric_logs[i]),
                             MetricLogger._extract_runtime_from_filename(original_metric_logs[i + 1])]) / 2

        nepochs = base_log.index.size
        first_fault_epoch = nepochs - n
        faulty_data = base_log.iloc[first_fault_epoch:]

        def corrupt_log(log: pd.DataFrame) -> pd.DataFrame:
            d1 = log.iloc[:first_fault_epoch]
            d2 = log.iloc[first_fault_epoch:]
            new_log = pd.concat([d1, faulty_data, d2], axis=0).reset_index(drop=True)
            new_log.index += 1
            return new_log

        source_of_corruption = corrupt_log(base_log)
        source_of_corruption.to_pickle(dtree.model_metrics_dir / f"{new_timestamp}.pkl.gz")

        for mlog in original_metric_logs[i + 1:]:
            original = MetricLogger._load_metrics_log(mlog, safe_load=True)
            corrupted = corrupt_log(original)
            assert corrupted.index.size - original.index.size == n, "Corruption did not work properly."
            corrupted.to_pickle(mlog)

    # introduce_fault(3, 3)

    backupdir = test_pth / "backup_dir"
    clean_corrupt_files(test_pth, 0, 1, cleanup=True, backupdir=backupdir)
    if corrupt:
        introduce_fault(1, 2)
    check_metric_data_integrity(backup_dir=backupdir, **tree_kwargs)





import logging
from enum import Enum

import yacs.config as config

_log = logging.getLogger(__name__)


class RegressionLossFuncTypes(Enum):
    squared_error = "reg:squarederror"
    custom = "reg:custom"


#  The default configuration for a squared error loss function.
default_se_loss_config = config.CfgNode()
default_se_loss_config.multioutput = "raw_values"

#  The default configuration for a custom exponential boundary loss function used as a
#  lower bound.
root = config.CfgNode()
root.y_lim = 0.
root.argmin = 0.5
root.c = None
root.min_val = 1e-6
root.multioutput = "raw_values"
default_exp_lower_bound_loss_config = root

# The default configuration for a custom exponential boundary loss function used as an
# upper bound.
root = config.CfgNode()
root.y_lim = 1.
root.argmin = 0.5
root.c = None
root.min_val = 1e-6
root.multioutput = "raw_values"
default_exp_upper_bound_loss_config = root


class default_base_loss_configs(Enum):
    exp_lower_bound = default_exp_lower_bound_loss_config
    exp_upper_bound = default_exp_upper_bound_loss_config
    se = default_se_loss_config


# The default configuration for a custom loss function that mixes multiple loss functions
# into a weighted average loss.
# 'funcs' and 'params' here correspond to lists of equal length containing the names of
# the loss functions and their corresponding keyword-arguments, respectively. The
# 'weights' are the mixture weights for the corresponding functions. The number of
# functions used here is flexible.
root = config.CfgNode()
root.funcs = ["exp_lower_bound", "exp_upper_bound", "se"]
root.params = [default_base_loss_configs[f].value.clone() for f in root.funcs]
root.weights = [1. / 3.] * 3
default_mixed_loss_config = root

# The default configuration for the transformations to be applied to the target variable
# This may be either a single transformation or a chain of transformations.
# In the case of a single transformation, set the key 'transform' to the name of the
# desired transformation and 'params' to the relevant keyword-arguments.
# In order to define a chain of transformations, set 'transform' to "chain" and pass
# a dictionary to 'params' containing two keys: "funcs" and "params", corresponding to
# lists of equal length containing the transformation names and keyword-arguments,
# respectively.
root = config.CfgNode()
root.transform = "chain"
root.params = config.CfgNode({
    "funcs": ["MinMax", "custom:inverse_sigmoid"],
    "params": [
        config.CfgNode({"feature_range": (1e-6, 1. - 1e-6)}),
        config.CfgNode({"k": 1.})
    ]
})
default_target_config = root

# The default configuration for a surrogate pipeline
# Regarding the loss function, instead of using a mixture of losses as here, a single
# loss function name can be passed to 'loss' and the corresponding keyword-arguments to
# 'loss_params'. 'loss_params' (or 'params' in the case of a mixture of losses) can also
# be the name of a .yaml file that resides in 'config_dir' in order to load the
# corresponding values from there.
default_pipeline_config = config.CfgNode()
default_pipeline_config.loss = RegressionLossFuncTypes.custom.value
default_pipeline_config.loss_params = default_mixed_loss_config.clone()
default_pipeline_config.config_dir = None
default_pipeline_config.target_config = default_target_config
default_pipeline_config.set_new_allowed(False)
default_pipeline_config.freeze()
del root



""" Very simple script for generating a reproducible set of hyper-parameters for the surrogate. """

from jahs_bench.tabular.surrogate import xgb
from sklearn.model_selection import ParameterSampler
import pickle
import numpy as np

# Randomly generated seed
rng = np.random.RandomState(312828812)
nsamples = 10
sampler = ParameterSampler(xgb.XGBSurrogate._hpo_search_space, n_iter=nsamples, random_state=rng)
params = list(sampler)
with open("/jahs_bench/tabular_sampling/surrogate/candidate_hps.json", "w") as fp:
    pickle.dump(params, fp)



"""
This script is intended to be used for resuming halted evaluations after the data from a partial run has been used to
pre-compute a better task distribution across multiple workers on a cluster.
"""


import argparse
import ConfigSpace
import logging
import json
from pathlib import Path
import sys

import pandas as pd
import torch
import traceback
from typing import Tuple, Sequence

import jahs_bench.tabular.lib.naslib.utils.logging as naslib_logging
import jahs_bench.tabular.lib.naslib.utils.utils as naslib_utils
from jahs_bench.tabular.lib.naslib.utils.utils import AttrDict

from jahs_bench.tabular.clusterlib import prescheduler as sched_utils
from jahs_bench.tabular.lib.core.constants import Datasets, standard_task_metrics, training_config
from jahs_bench.tabular.lib.core import utils
from jahs_bench.tabular.lib.core import datasets as dataset_lib
from jahs_bench.tabular.lib.core.procs import train
from jahs_bench.tabular.search_space import NASB201HPOSearchSpace


def argument_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument("--rootdir", type=Path, default=Path().cwd(),
                        help="Path to the root directory where all the tasks' output will be stored. Task-specific "
                             "sub-directories will be created here if needed. Note that this is NOT the same as the "
                             "'basedir' of a DirectoryTree object as multiple DirectoryTree objects will be created "
                             "as and when needed.")
    parser.add_argument("--workerid", type=int,
                        help="An offset from 0 for this worker within the current job's allocation of workers. This "
                             "does NOT correspond to a single 'taskid' but rather the portfolio of configurations that "
                             "this worker will evaluate.")
    parser.add_argument("--workerid_offset", type=int, default=0,
                        help="An additional fixed offset from 0 for this worker that, combined with the value of "
                             "'--worker', gives the overall worker ID in the context of a job that has been split into "
                             "multiple smaller parts.")
    parser.add_argument("--datadir", type=Path, default=dataset_lib.get_default_datadir(),
                        help="The directory where all datasets are expected to be stored.")
    parser.add_argument("--portfolio", type=Path, default=None,
                        help="Path to a pickled Pandas Series that contains the portfolio for this job's workers.")
    parser.add_argument("--debug", action="store_true", help="Enable debug mode (very verbose) logging.")

    # Unpack the training config overrides into CLI flags.
    # for k, v in recognized_train_config_overrides.items():
    for k, v in training_config.items():
        parser.add_argument(f"--{k}", **v)

    return parser


def parse_training_overrides(args: argparse.Namespace) -> AttrDict:
    return AttrDict({k: getattr(args, k) for k in recognized_train_config_overrides.keys()})


def get_tranining_config_from_args(args: Sequence[str]) -> AttrDict:
    return AttrDict({k: getattr(args, k) for k in training_config.keys()})


def reload_train_config(dtree: utils.DirectoryTree, **overrides) -> Tuple[AttrDict, Datasets]:
    """ Load a saved training config from disk, overriding the training config with appropriate values if needed. """

    with open(dtree.task_config_file, "r") as fp:
        task_config = json.load(fp)

    task_config["train_config"] = {**task_config["train_config"],
                                   **{k: v for k, v in overrides.items() if v is not None}}

    try:
        dataset = [v for v in Datasets.__members__.values() if v.value[0] == task_config["dataset"]][0]
    except IndexError:
        raise RuntimeError(f"Unable to determine appropriate dataset with name - {task_config['dataset']}.")

    # Update the saved task config with the latest overrides.
    ## DISABLED since this can lead to race conditions during multiprocess write operations
    # with open(dtree.task_config_file, "w") as fp:
    #     json.dump(task_config, fp, default=str)

    return AttrDict(task_config["train_config"]), dataset


def load_model_config(task_metrics: AttrDict, model_idx: int) -> Tuple[int, dict]:
    """ Given the metrics of a task, recover the specified model's configuration. Remember that 'model_idx' uses a
    starting index of 1 and not 0. """

    # n_known_samples = len(task_metrics.model_idx)
    # if n_known_samples < model_idx:
    try:
        global_seed = task_metrics.global_seed[model_idx - 1]
        model_config = task_metrics.model_config[model_idx - 1]
    except IndexError as e:
        raise IndexError(f"The specified model index {model_idx} was not found in the task's recorded metrics, which "
                         f"contained {len(task_metrics.model_idx)} records.") from e

    return global_seed, model_config


def instantiate_model(model_config: dict, dataset: Datasets) -> NASB201HPOSearchSpace:
    """ Given a model configuration dict and a dataset, instantiate a model with the corresponding configuration. """

    NASB201HPOSearchSpace.NUM_CLASSES = dataset.value[3]
    search_space = NASB201HPOSearchSpace()
    if dataset.value[2] == 1:  # Check # of channels
        # The dataset is in Grayscale
        search_space.GRAYSCALE = True

    assert utils.adapt_search_space(search_space, portfolio=None, taskid=None, opts=model_config, suffix=None), \
        f"Unable to restrict the search space to use the specified model configuration:\n{model_config}"
    assert all([isinstance(h, ConfigSpace.Constant) for h in search_space.config_space.get_hyperparameters()]), \
        f"Failed to restrict all hyperparameters in the search space to the specified model configuration. The " \
        f"restricted space is defined by:\n{search_space.config_space}"

    # Since all hyper-parameters are constants, this will always sample the same configuration.
    search_space.sample_random_architecture()

    return search_space

# TODO: Extend this script to also include stage 2 verification functionality - loading a specific checkpoint, training
#  it for a specified number of epochs, and verifying the metric data. Save the new metrics as alternative data points
#  for a potential analysis of the various metric distributions
def resume_work(basedir: Path, taskid: int, model_idx: int, datadir: Path, debug: bool = False,
                logger: logging.Logger = None, **training_config_overrides):
    """ Resume working on a particular configuration wherever it was left off by loading all relevant parameters from
    the saved checkpoints. Some training parameters may be overridden by providing the relevant overrides as keyword
    arguments. """

    dir_tree = utils.DirectoryTree(basedir=basedir, taskid=taskid, model_idx=model_idx)
    if logger is None:
        logger = naslib_logging.setup_logger(str(dir_tree.model_dir / f"resume.log"))

    task_metrics = AttrDict(utils.attrdict_factory(metrics=standard_task_metrics, template=list))
    # No timer, since this is read-only - reads the pre-sampled configs from disk
    _ = utils.MetricLogger(dir_tree=dir_tree, metrics=task_metrics, timer=None,
                           set_type=utils.MetricLogger.MetricSet.task, logger=logger)

    try:
        global_seed, model_config = load_model_config(task_metrics, model_idx)
    except IndexError:
        logger.warning(f"Could not resume working on model config of task {taskid}, model index {model_idx} due to an "
                       f"error:", exc_info=True)
        return

    try:
        train_config, dataset = reload_train_config(dir_tree, **training_config_overrides)
    except RuntimeError:
        logger.warning(f"Could not resume working on model config of task {taskid}, model index {model_idx} due to an "
                       f"error:", exc_info=True)
        return

    logger.setLevel(logging.DEBUG if debug else logging.INFO)

    ## Model initialization
    model = instantiate_model(model_config, dataset)
    model.parse()
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    logger.debug(f"Training model {model_idx} on device {str(device)}")

    transfer_devices = device.type != "cpu"
    if transfer_devices:
        model = model.to(device)

    try:
        naslib_utils.set_seed(global_seed)

        data_loaders, min_shape = dataset_lib.get_dataloaders(
            dataset=dataset, batch_size=train_config.batch_size, cutout=0, split=train_config.split,
            resolution=model_config.get("Resolution", 0), trivial_augment=model_config.get("TrivialAugment", False),
            datadir=datadir
        )
        validate = "valid" in data_loaders
        train(
            model=model, data_loaders=data_loaders, train_config=train_config, dir_tree=dir_tree, logger=logger,
            validate=validate, transfer_devices=transfer_devices, device=device, debug=debug, min_shape=min_shape
        )

    except Exception as e:
        logger.info("Architecture Training failed.", exc_info=True)
        error_description = {
            "exception": traceback.format_exc(),
            "config": model_config,
            "global_seed": global_seed
        }
        with open(dir_tree.model_error_description_file, "w") as fp:
            json.dump(error_description, fp, indent=4)
        if debug:
            raise e
    else:
        if dir_tree.model_error_description_file.exists():
            dir_tree.model_error_description_file.unlink()
        del model  # Release memory
        dir_tree.model_idx = None  # Ensure that previously written data cannot be overwritten

    ## Clean up memory before terminating task
    del task_metrics
    del dir_tree


if __name__ == "__main__":
    ## Parse CLI
    args = argument_parser().parse_args()

    train_config_overrides = get_tranining_config_from_args(args)

    workerid = args.workerid + args.workerid_offset

    logdir: Path = args.rootdir / sched_utils.logdir_name
    logdir.mkdir(exist_ok=True, parents=False)
    logger = naslib_logging.setup_logger(str(logdir / f"{workerid}.log"), name='tabular_sampling')
    if args.debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.WARNING)

    portfolio = pd.read_pickle(args.portfolio)
    if "worker_id" not in portfolio.index.names:
        raise ValueError(f"The given portfolio at {args.portfolio} has not been configured properly, its index lacks "
                         f"the level for a worker ID. Portfolio index level names: {portfolio.index.names}")
    if workerid not in portfolio.index.unique("worker_id"):
        logger.warning(f"Worker {workerid} not present in the job's portfolio. Ignore this warning if the job "
                       f"specification had more workers than needed, otherwise, check the portfolio file "
                       f"{args.portfolio}")
        sys.exit(1)

    portfolio = portfolio.xs(workerid, level="worker_id")
    # TODO: Remove worker-specific portfolio_dir and relative functionality once it is confirmed to be superfluous
    worker_config = sched_utils.WorkerConfig(worker_id=workerid, portfolio=portfolio)

    for taskid, model_idx, basedir in worker_config.iterate_portfolio(rootdir=args.rootdir):
        resume_work(basedir, taskid, model_idx, datadir=args.datadir, debug=args.debug, logger=logger,
                    **train_config_overrides)




import os
import pickle

import ConfigSpace
import numpy as np
import random
import torch
import torch.nn as nn

from jahs_bench.tabular.lib.naslib.search_spaces.core import primitives as core_ops
from jahs_bench.tabular.lib.naslib.search_spaces.core.graph import Graph, EdgeData
from jahs_bench.tabular.lib.naslib.search_spaces.core.primitives import AbstractPrimitive
from jahs_bench.tabular.lib.naslib.search_spaces.core.query_metrics import Metric

from jahs_bench.tabular.lib.naslib.utils.utils import get_project_root

from .conversions import convert_op_indices_to_naslib, convert_naslib_to_op_indices, convert_naslib_to_str
from .primitives import ResNetBasicblock, ConvBN, StemGrayscale
from .configspace import joint_config_space
from .constants import OP_NAMES, Activations


class NASB201HPOSearchSpace(Graph):
    """
    Implementation of the nasbench 201 + HPO search space.
    """

    OPTIMIZER_SCOPE = [
        "stage_1",
        "stage_2",
        "stage_3",
    ]

    QUERYABLE = False

    # When this is True, newly created graphs (class instances) are immediately compiled into PyTorch models
    COMPILE_ON_CREATE = True

    GRAYSCALE = False # TODO: Verify implementation
    NUM_CELL_STAGES = 3 # Cannot be less than 1, abbreviated as S TODO: Finish and verify implementation
    KNOWN_CHANNEL_WIDTHS = {
        1: (4, 8, 16),
        2: (8, 16, 32),
        3: (16, 32, 64),
    }

    config_space = joint_config_space
    _config = joint_config_space.get_default_configuration()

    def __init__(self):
        # We re-define the way SearchSpaces are typically initialized by NASLib in order to accommodate for the greater
        # dynamic flexibility in graph design required by HPO, which may include fidelity parameters. Thus, the
        # initializer now only sets a number of fixed parameters that provide some general information about the
        # search space and calls a separate function to instantiate a graph. Thus, the graph can be cleared and
        # re-constructed using different parameters any time by calling the relevant function again.

        super().__init__()
        self.num_classes = self.NUM_CLASSES if hasattr(self, 'NUM_CLASSES') else 10
        self.op_indices = None

        self.space_name = 'nasbench201_hpo'
        self._construct_graph()

    def _construct_graph(self):
        assert self.NUM_CELL_STAGES >= 1, f"The search space must have at least one cell stage but " \
                                          f"{self.NUM_CELL_STAGES} stages were specified."

        cell_repeat = self.config.get("N")
        # channels = self.KNOWN_CHANNEL_WIDTHS[self.config.get("W")]
        channels = tuple(self.config.get("W") * (2 ** i) for i in range(3))
        activation = self.config.get("Activation")

        #
        # Cell definition
        #
        cell = Graph()
        cell.name = "cell"  # Use the same name for all cells with shared attributes

        # Input node
        cell.add_node(1)

        # Intermediate nodes
        cell.add_node(2)
        cell.add_node(3)

        # Output node
        cell.add_node(4)

        # Edges
        cell.add_edges_densly()

        #
        # Makrograph definition
        #
        self.name = "makrograph"

        # Cell is on the edges
        # 1-2:                                      Preprocessing
        # 2-3, ..., (2+N-1)-(2+N):                  cells stage 1
        #
        # (2+N)-(3+N):                              residual block stride 2
        # (3+N)-(3+N+1), ..., (3+2*N-1)-(3+2*N):    cells stage 2
        #
        # (3+2*N)-(4+2*N):                          residual block stride 2
        # (4+2*N)-(4+2*N+1), ..., (4+3*N-1)-(4+3*N): cells stage 3
        # .
        # .
        # .
        # (1+S*(N+1))-(1+S*(N+1)+1):                          post-processing

        total_num_nodes = 1 + self.NUM_CELL_STAGES * (1 + cell_repeat) + 1
        self.add_nodes_from(range(1, total_num_nodes + 1))
        self.add_edges_from([(i, i + 1) for i in range(1, total_num_nodes)])

        edge_names = {}

        edge_names["preproc"] = (1, 2)
        edge_names["stage1"] = [(i, i + 1) for i in range(2, 2 + cell_repeat)]

        # TODO: Make the number of stages programmatically dynamic
        edge_names["res1"] = (edge_names["stage1"][-1][-1], edge_names["stage1"][-1][-1] + 1)
        edge_names["stage2"] = [(i, i + 1) for i in range(edge_names["res1"][1], edge_names["res1"][1] + cell_repeat)]
        edge_names["res2"] = (edge_names["stage2"][-1][1], edge_names["stage2"][-1][1] + 1)
        edge_names["stage3"] = [(i, i + 1) for i in range(edge_names["res2"][1], edge_names["res2"][1] + cell_repeat)]
        edge_names["postproc"] = (edge_names["stage3"][-1][1], edge_names["stage3"][-1][1] + 1)

        #
        # operations at the edges
        #

        # preprocessing
        self.edges[edge_names["preproc"]].set(
            'op', StemGrayscale(channels[0]) if self.GRAYSCALE else core_ops.Stem(channels[0]))

        # stage 1
        for e in edge_names["stage1"]:
            self.edges[e].set('op', cell.copy().set_scope('stage_1'))

        # stage 2
        self.edges[edge_names["res1"]].set('op', ResNetBasicblock(C_in=channels[0], C_out=channels[1], stride=2,
                                                                  activation=activation))
        for e in edge_names["stage2"]:
            self.edges[e].set('op', cell.copy().set_scope('stage_2'))

        # stage 3
        self.edges[edge_names["res2"]].set('op', ResNetBasicblock(C_in=channels[1], C_out=channels[2], stride=2,
                                                                  activation=activation))
        for e in edge_names["stage3"]:
            self.edges[e].set('op', cell.copy().set_scope('stage_3'))

        # post-processing
        self.edges[edge_names["postproc"]].set('op', core_ops.Sequential(
            nn.BatchNorm2d(channels[-1]),
            # nn.SiLU(inplace=False) if use_swish else nn.ReLU(inplace=False),
            Activations.__members__[activation].value[1](inplace=False),
            nn.AdaptiveAvgPool2d(1),
            nn.Flatten(),
            nn.Linear(channels[-1], self.num_classes)
        ))

        self.edge_names = edge_names

        # set the ops at the cells (channel dependent)
        for c, scope in zip(channels, self.OPTIMIZER_SCOPE):
            self.update_edges(
                update_func=lambda edge: _set_cell_ops(edge, C=c, activation=activation),
                scope=scope,
                private_edge_data=True
            )

        op_indices = [self.config.get(f"Op{i}") for i in range(1, 7)]
        self.set_op_indices(op_indices)
        if self.COMPILE_ON_CREATE:
            self.compile()

    @property
    def config(self) -> ConfigSpace.Configuration:
        return self._config

    @config.setter
    def config(self, val: ConfigSpace.Configuration):
        try:
            self.config_space.check_configuration(val)
        except Exception as e:
            raise RuntimeError(f"The given config value could not be verified against the configuration space."
                               f"\nConfiguration space: {str(self.config_space)}\nGiven config: {str(val)}") \
                from e
        else:
            self._config = val

    def get_op_indices(self):
        if self.op_indices is None:
            self.op_indices = convert_naslib_to_op_indices(self)
        return self.op_indices

    def get_hash(self):
        return tuple(self.config.get_dictionary().items())

    def set_op_indices(self, op_indices):
        # This will update the edges in the naslib object to op_indices
        self.op_indices = op_indices
        convert_op_indices_to_naslib(op_indices, self)

    def sample_random_architecture(self, dataset_api=None, rng=None):
        """
        This will sample a random architecture from the search space. Sampling here means that first a fidelity will be
        sampled from the fidelity space, a new graph will be constructed, random indices for the graph will be sampled
        and finally edges in the naslib object will be updated accordingly. Fidelity sampling can be turned off by
        setting sample_fidelity to False (default: True). If fidelity sampling is off, the naslib object's fidelity
        attribute must be set to a valid fidelity value before this function is called. 'rng' can be either None, in
        which case a fresh random number generator is used, or an integer value used to seed a new instance of
        numpy.random.default_rng() or any compatible random state for the same, including an existing RNG to ensure
        consistent, reproducible results.
        """
        self.clear()
        if not isinstance(rng, np.random.RandomState):
            # Assume that rng is either None or a compatible source of entropy
            rng = np.random.RandomState(rng)
        self.config_space.random = rng
        self.config = self.config_space.sample_configuration()
        self._construct_graph()

    def get_type(self):
        return 'nasbench201_hpo'


def _set_cell_ops(edge, C, activation=Activations.ReLU.name):
    edge.data.set('op', [
        core_ops.Identity(),
        core_ops.Zero(stride=1),
        ConvBN(C, C, kernel_size=3, activation=activation),
        ConvBN(C, C, kernel_size=1, activation=activation),
        core_ops.AvgPool1x1(kernel_size=3, stride=1),
    ])



import argparse
import functools
import logging
from pathlib import Path
from typing import Tuple, Optional

import joblib
import neps
import pandas as pd
import scipy.stats
from jahs_bench.surrogate import model, config as cfg

_log = logging.getLogger(__name__)

xgb_hp_space = {
    "max_depth": neps.IntegerParameter(
        lower=1, upper=15, default=6, default_confidence="low"
    ),
    "min_child_weight": neps.IntegerParameter(
        lower=1, upper=10, default=1, default_confidence="low"
    ),
    "colsample_bytree": neps.FloatParameter(
        lower=0., upper=1., log=False, default=1., default_confidence="low"
    ),
    "learning_rate": neps.FloatParameter(
        lower=0.001, upper=0.5, log=True, default=0.3, default_confidence="low"
    ),
    "colsample_bylevel":  neps.FloatParameter(
        lower=0., upper=1., log=False, default=1., default_confidence="low"
    ),
    "sigmoid_k": neps.FloatParameter(
        lower=0.01, upper=2., log=True, default=1., default_confidence="low"
    )
}

def train_surrogate(working_directory: Path, train_data: pd.DataFrame,
                    valid_data: pd.DataFrame, **config_dict):
    """

    :param working_directory:
    :param train_data: pandas DataFrame
        A pandas DataFrame object with 2-level MultiIndex columns, containing the
        training data. Level 0 should contain the columns "features" and "labels". Level
        1 can contain arbitrary columns.
    :param valid_data: pandas DataFrame
        A pandas DataFrame object with 2-level MultiIndex columns, containing the
        validation data. Level 0 should contain the columns "features" and "labels".
        Level 1 can contain arbitrary columns, but they should match those in
        `train_data`.
    :param config_dict: keyword-arguments
        These specify the various hyperparameters to be used for the model.
    :return:
    """

    xtrain = train_data["features"]
    ytrain = train_data["labels"]

    xvalid = valid_data["features"]
    yvalid = valid_data["labels"]

    _log.info(f"Preparing to train surrogate.")
    pipeline_config = cfg.default_pipeline_config.clone()
    xgb_params = config_dict.copy()
    sigmoid_k = xgb_params.pop("sigmoid_k", None)

    if sigmoid_k is not None:
        pipeline_config.defrost()
        pipeline_config.target_config.params.params[1].k = float(sigmoid_k)
        pipeline_config.freeze()

    surrogate = model.XGBSurrogate(hyperparams=xgb_params, use_gpu=True)

    _log.info("Training surrogate.")
    random_state = None
    scores = surrogate.fit(xtrain, ytrain, random_state=random_state,
                           pipeline_config=pipeline_config)
    _log.info(f"Trained surrogate has scores: {scores}")

    modeldir = working_directory / "xgb_model"
    _log.info(f"Saving trained surrogate to disk at {str(modeldir)}")
    modeldir.mkdir()
    surrogate.dump(modeldir)
    config_file = modeldir / "pipeline_config.pkl.gz"
    joblib.dump(pipeline_config, config_file, protocol=4)

    _log.info(f"Generating validation scores.")
    ypred = surrogate.predict(xvalid)
    valid_score = scipy.stats.kendalltau(yvalid, ypred)

    _log.info(f"Trained surrogate has validation score: {valid_score}")
    # Return negative KT correlation since NEPS minimizes the loss
    return -valid_score.correlation

def load_data(datadir: Path, output: Optional[str]) -> Tuple[pd.DataFrame, pd.DataFrame]:
    _log.info("Loading training data.")
    train_data: pd.DataFrame = pd.read_pickle(datadir / "train_set.pkl.gz")
    _log.info(f"Loaded training data of shape {train_data.shape}")

    _log.info("Loading validation data.")
    valid_data: pd.DataFrame = pd.read_pickle(datadir / "valid_set.pkl.gz")
    _log.info(f"Loaded validation data of shape {valid_data.shape}")

    features = train_data["features"].columns
    labels = train_data["labels"].columns

    assert features.difference(valid_data["features"].columns).size == 0, \
        "Mismatch between the input features in the training and validation sets."
    assert labels.difference(valid_data["labels"].columns).size == 0, \
        "Mismatch between the output labels in the training and validation sets."

    if output is not None:
        assert output in labels, f"The chosen prediction label {output} is not present " \
                                 f"in the training data."
        selected_cols = train_data[["features"]].columns.tolist() + [("labels", output)]
        train_data = train_data.loc[:, selected_cols]
        valid_data = valid_data.loc[:, selected_cols]

    return train_data, valid_data

def perform_hpo(working_directory: Path, datadir: Path, output: str,
                max_evaluations_total: int = 5,):
    _log.info(f"Performing HPO using the working directory {working_directory}, using "
              f"the data at {datadir}, over {max_evaluations_total} evaluations.")

    train_data, valid_data = load_data(datadir=datadir, output=output)
    pipeline = functools.partial(train_surrogate, train_data=train_data,
                                 valid_data=valid_data)
    neps.run(
        run_pipeline=pipeline,
        pipeline_space=xgb_hp_space,
        working_directory=working_directory,
        max_evaluations_total=max_evaluations_total,
    )

    _log.info("Finished.")

def parse_cli():
    parser = argparse.ArgumentParser(
        "Use the NEPS framework to train a surrogate model by iteratively optimizing "
        "the hyperparameters."
    )
    parser.add_argument("--working_directory", type=Path,
                        help="A working directory to be used by NEPS for various tasks "
                             "that require writing to disk.")
    parser.add_argument("--datadir", type=Path,
                        help="Path to the directory where the cleaned, tidy training "
                             "and validation data splits to be used are stored.")
    parser.add_argument("--output", type=str,
                        help="Which of the available performance metrics to train the "
                             "surrogate for.")
    parser.add_argument("--max_evaluations_total", type=int, default=5,
                        help="Number of evaluations that this NEPS worker should "
                             "perform.")

    args = parser.parse_args()
    return args

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO,
                        format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
                        datefmt="%m/%d %H:%M:%S")
    _log.setLevel(logging.INFO)
    args = parse_cli()
    perform_hpo(**vars(args))



# Hard-code the import order to easily detect cyclic imports
from . import constants
from . import configspace
from . import conversions
from . import primitives
from . import graph

from .graph import NASB201HPOSearchSpace




"""
There are three representations
'naslib': the NASBench201SearchSpace object
'op_indices': A list of six ints, which is the simplest representation
'arch_str': The string representation used in the original nasbench201 paper

This file currently has the following conversions:
naslib -> op_indices
op_indices -> naslib
naslib -> arch_str

Note: we could add more conversions, but this is all we need for now
"""

import ConfigSpace
from .constants import OP_NAMES, EDGE_LIST, nb201_to_ops, ops_to_nb201

def convert_naslib_to_op_indices(naslib_object):

    cell = naslib_object._get_child_graphs(single_instances=True)[0]
    ops = []
    for i, j in EDGE_LIST:
        ops.append(cell.edges[i, j]['op'].get_op_name)

    return [OP_NAMES.index(name) for name in ops]


def convert_op_indices_to_naslib(op_indices, naslib_object):
    """
    Converts op indices to a naslib object
    input: op_indices (list of six ints)
    naslib_object is an empty NasBench201SearchSpace() object.
    Do not call this method with a naslib object that has already been 
    discretized (i.e., all edges have a single op).

    output: none, but the naslib object now has all edges set
    as in genotype.
    
    warning: this method will modify the edges in naslib_object.
    """
    
    # create a dictionary of edges to ops
    edge_op_dict = {}
    for i, index in enumerate(op_indices):
        edge_op_dict[EDGE_LIST[i]] = OP_NAMES[index]
    
    def add_op_index(edge):
        # function that adds the op index from the dictionary to each edge
        if (edge.head, edge.tail) in edge_op_dict:
            for i, op in enumerate(edge.data.op):
                if op.get_op_name == edge_op_dict[(edge.head, edge.tail)]:
                    index = i
                    break
            edge.data.set('op_index', index, shared=True)

    def update_ops(edge):
        # function that replaces the primitive ops at the edges with the one in op_index
        if isinstance(edge.data.op, list):
            primitives = edge.data.op
        else:
            primitives = edge.data.primitives

        edge.data.set('op', primitives[edge.data.op_index])
        edge.data.set('primitives', primitives)     # store for later use

    naslib_object.update_edges(
        add_op_index,
        scope=naslib_object.OPTIMIZER_SCOPE,
        private_edge_data=False
    )
    
    naslib_object.update_edges(
        update_ops, 
        scope=naslib_object.OPTIMIZER_SCOPE,
        private_edge_data=True
    )

def convert_naslib_to_str(naslib_object):
    """
    Converts naslib object to string representation.
    """

    cell = naslib_object.edges[2, 3].op
    edge_op_dict = {
        (i, j): ops_to_nb201[cell.edges[i, j]['op'].get_op_name] for i, j in cell.edges
    }
    op_edge_list = [
        '{}~{}'.format(edge_op_dict[(i, j)], i-1) for i, j in sorted(edge_op_dict, key=lambda x: x[1])
    ]

    return '|{}|+|{}|{}|+|{}|{}|{}|'.format(*op_edge_list)


def convert_str_to_indices(arch_str: str) -> list:
    """ Converts a given string denoting the cell architecture in the format used by the original NASBench-201 data
    into a list of indices that can be used by NASLib NASB201HPOSearchSpace objects. """

    nodestrs = arch_str.split('+')
    edge_ops = {k: None for k in EDGE_LIST}
    out_node = 2
    for i, node_str in enumerate(nodestrs):
        inputs = list(filter(lambda x: x != '', node_str.split('|')))
        for xinput in inputs: assert len(xinput.split('~')) == 2, 'invalid input length : {:}'.format(xinput)
        inputs = (xi.split('~') for xi in inputs)
        try:
            input_infos = tuple((nb201_to_ops[nb201_op], int(IDX) + 1) for (nb201_op, IDX) in inputs)
        except KeyError as e:
            raise RuntimeError("Failed to translate the given NB201 architecture string to a NASLib architecture "
                               "description.") from e
        for op, in_node in input_infos:
            edge_name = in_node, out_node
            if edge_name not in edge_ops:
                raise ValueError(f"The given architecture string requires an edge at {edge_name}, which is "
                                 f"an invalid cell configuration.")
            edge_ops[edge_name] = OP_NAMES.index(op)
        out_node += 1

    return tuple(edge_ops[k] for k in EDGE_LIST)

def convert_nb201_str_to_naslib_obj(nb201_str, naslib_obj, change_search_space=False):
    """ Given a string describing a unique architecture in the format of NASBench-201, changes the given NASLib Object,
    expected to be a NASB201HPOSearchSpace object, to reflect that architecture. Note that this function assumes that
    the other dynamic architecture parameters of the object's config have already been properly set. Alternatively,
    setting change_search_space to True will instead cause this function to modify the ConfigurationSpace of naslib_obj
    such that the cell ops are fixed to the appropriate values for all future samples drawn from the space. This still
    does not touch the other parameters of the space. Ideally, this setting should be used on the entire
    NASB201HPOSearchSpace class. """

    op_inds = convert_str_to_indices(nb201_str)

    if change_search_space:
        original_space = naslib_obj
        config_space = original_space.config_space
        known_params = {p.name: p for p in config_space.get_hyperparameters()}
        new_params = {f"Op{i}": v for i, v in enumerate(op_inds, start=1)}
        for k, v in new_params.items():
            meta = dict(known_params[k].meta, **dict(constant_overwrite=True))
            known_params[k] = ConfigSpace.Constant(k, v, meta=meta)
        new_config_space = ConfigSpace.ConfigurationSpace(f"{config_space.name}_fixed_ops")
        new_config_space.add_hyperparameters(known_params.values())

        original_space.config_space = new_config_space
    else:
        naslib_obj.clear()
        config = naslib_obj.config_space.get_default_configuration()
        config.keys()
        op_config = {f"Op{i}": op for i, op in enumerate(op_inds, start=1)}
        for k, v in op_config.items():
            config[k] = v

        naslib_obj.config = config
        naslib_obj._construct_graph()



"""
Main script for sampling from a specified joint NAS+HPO search space, training and evaluating the sampled
configurations and recording the metrics when run on a server. Intended for use with a SLURM script such that one job
launches multiple parallel python processes (tasks), each running any specified number of logical CPUs. Each task then
evaluates a number of configurations (models). The data is generated and stored using the following directory structure
template:
<base_dir>    <-- Possibly a job-level directory
|    |--> <task_idx>/
|    |    |--> metrics/
|    |    |    |--> <task-level metrics, if any>
|    |    |--> models /
|    |    |--> <model_idx>/
|    |    |    |--> metrics/
|    |    |    |    |--> <model-level metrics, if any>
|    |    |    |--> checkpoints/
|    |    |    |    |--> <model checkpoints, if any>

The PRNGs are setup in such a way that, for the hard-coded, fixed seed value, the same task, as identified by an
integer task_id, will generate the same sequence of samples from any given search space as well as the same sequence of
global seeds for numpy/torch/random.

"""

import argparse
import json
import logging
import os
import traceback
from pathlib import Path
from typing import Iterable, Sequence, Optional, Union

import numpy as np
import torch

import jahs_bench.tabular.lib.naslib.utils.logging as naslib_logging
import jahs_bench.tabular.lib.naslib.utils.utils as naslib_utils
from jahs_bench.tabular.lib.naslib.utils.utils import AttrDict
from jahs_bench.tabular.lib.core.constants import Datasets, standard_task_metrics
from jahs_bench.tabular.lib.core.constants import training_config as _training_config
from jahs_bench.tabular.lib.core import utils
from jahs_bench.tabular.lib.core import datasets as dataset_lib
from jahs_bench.tabular.lib.core.procs import train
from jahs_bench.tabular.search_space import NASB201HPOSearchSpace

# Randomly generated entropy source, to remain fixed across experiments.
_seed = 79029434164686768057103648623012072794


def get_tranining_config_from_args(args: Sequence[str]) -> AttrDict:
    return AttrDict({k: getattr(args, k) for k in _training_config.keys()})


def argument_parser():
    parser = argparse.ArgumentParser()

    # Unpack the training config into CLI flags.
    for k, v in _training_config.items():
        parser.add_argument(f"--{k}", **v)

    parser.add_argument("--basedir", type=Path, default=Path().cwd(),
                        help="Path to the base directory where all the tasks' output will be stored. Task-specific "
                             "sub-directories will be created here if needed.")
    parser.add_argument("--datadir", type=Path, default=dataset_lib.get_default_datadir(),
                        help="The directory where all datasets are expected to be stored.")
    parser.add_argument("--taskid", type=int,
                        help="An offset from 0 for this task within the current node's allocation of tasks.")
    parser.add_argument("--taskid_base", type=int, default=0,
                        help="An additional offset from 0 to manually shift all task IDs further, useful when sampling "
                             "is intended to continue adding more data beyond a previous job's data.")
    parser.add_argument("--dataset", type=str, choices=list(Datasets.__members__.keys()),
                        help="The name of which dataset is to be used for model training and evaluation. Only one of "
                             "the provided choices can be used.")
    parser.add_argument("--nsamples", type=int, default=100,
                        help="Sets the maximum number of samples to be drawn from the search space for each task. Use "
                             "-1 to set this to unlimited.")
    parser.add_argument("--portfolio", type=Path, default=None,
                        help="Path to a saved pandas DataFrame or Series containing a portfolio of configurations. If "
                             "given, the portfolio is read and used to restrict the search space on a per-task basis. "
                             "Key-value pairs specified as CLI arguments for 'opts' override corresponding portfolio "
                             "values.")
    parser.add_argument("--cycle_portfolio", action="store_true",
                        help="Only appliable when a complete portfolio of task- and mode-wise configurations is given. "
                             "When given, causes the portfolio configurations within each task to cycle infinitely or "
                             "until sampling is stopped due to the limit set down by '--nsamples'.")
    parser.add_argument("--global_seed", type=int, default=None,
                        help="A value for a global seed to be used for all global NumPy and PyTorch random "
                             "operations. This is different from the fixed seed used for reproducible search space "
                             "sampling. If not specified, a random source of entropy is used instead.")
    # parser.add_argument("--standalone-mode", action="store_true",
    #                     help="Switch that enables working in a single-task local setup as opposed to the default "
    #                          "multi-node cluster setup.")
    parser.add_argument("--debug", action="store_true", help="Enable debug mode (very verbose) logging.")
    parser.add_argument("--generate_sampling_profile", action="store_true",
                        help="When given, does not actually train the sampled models. Instead, samples are simply "
                             "repeatedly drawn at 1s intervals and saved in order to build a profile of expected "
                             "samples.")
    parser.add_argument("opts", nargs=argparse.REMAINDER, default=None,
                        help="A variable number of optional keyword arguments provided as 2-tuples, each potentially "
                             "corresponding to a hyper-parameter in the search space. If a match is found, that "
                             "hyper-parameter is excluded from the search space and fixed to the given value instead.")
    return parser


def run_task(basedir: Path, taskid: int, train_config: AttrDict, dataset: Datasets, datadir,
             local_seed: Optional[int] = None, global_seed: Optional[Union[Iterable[int], int]] = None,
             debug: bool = False, generate_sampling_profile: bool = False, nsamples: int = 0,
             portfolio_pth: Optional[Path] = None, cycle_portfolio: bool = False, logger: logging.Logger = None,
             opts: Optional[Sequence[str]] = None):
    """
    Run the sampling, training and evaluation procedures on a single task.

    :param basedir: Path-like
        The base directory in which the directory tree will be created
    :param taskid: int
        The task ID of the task to be run.
    :param train_config: AttrDict
        A dictionary containing various key-value pairs describing the per-model training configuration. Consult
        'get_training_config_help()' and 'get_training_config_from_args()' for more details.
    :param dataset: str
        The name of a known dataset to be used for model training and evaluation.
    :param datadir: Path-like
        The path to a directory where all the datasets are stored.
    :param local_seed: int
        An integer seed for a locally instantiated and controlled RNG that will be passed around to control this task's
        sampling processes. If None (default), local system entropy is used as a seed instead. Note that for
        reproducibility, a fixed local seed should be passed.
    :param global_seed: Iterable over ints or int, optional
        Used for maintaining reproducibility across model training. Could be either a single integer, used to seed the
        global RNGs of PyTorch, NumPy and Random for every sampled model with the same seed, or an iterable that
        generates a global seed for each sampled model. If None (default), the local RNG is used to automatically
        sample a reproducible sequence of global seeds for each model. Consult 'default_global_seed_gen()' for more
        details.
    :param debug: bool
        A flag that, when set, enables debug mode logging and operation. Default: False.
    :param generate_sampling_profile: bool
        A flag that, when set, indicates that only a "dry run" of the sampling procedure should be done in order to
        generate the expected sampling profile of this task without actually training/evaluating the sampled models.
        Also consult 'nsamples'. Default: False.
    :param nsamples: int
        The maximum number of samples to be drawn from the search space for each task. When set to -1, draws an
        unlimited number of samples. Default: 0.
    :param portfolio_pth: Path-like
        Path to a '.pkl.gz' file containing a portfolio of configurations in the form of a compatible Pandas DataFrame.
        If only task-level configurations are given, the search space is restricted on a task-basis such that the given
        task's sampled configurations always conform to the portfolio's specifications. If both task-level and
        model-level configurations are given, the search space isn't sampled at all and instead the precise
        configurations specified in the portfolio are used to generate models. Also see 'cycle_portfolio'.
    :param cycle_portfolio: bool
        Only applicable when a portfolio is provided (see 'portfolio_pth'). When True, the configurations of each task
        are cycled infinitely or until the limit specified by 'nsamples' is reached. Default: False
    :param opts: Sequence of str
        A sequence of strings read as pairs of values that modify the search space such that the first string in each
        pair corresponds to a known parameter in the relevant config space and the second string corresponds to a
        constant value for that parameter, interpreted and type-cast according to the properties of the parameter in
        the defined config space. Default: None.
    :return:
    """

    rng = np.random.RandomState(np.random.Philox(seed=local_seed, counter=taskid))
    global_seed_gen = utils.default_global_seed_gen(rng, global_seed)
    dir_tree = utils.DirectoryTree(basedir=basedir, taskid=taskid)

    if logger is None:
        logger = naslib_logging.setup_logger(str(dir_tree.task_dir / "log.log"))

    task_metrics = AttrDict(
        utils.attrdict_factory(metrics=standard_task_metrics, template=list))

    # The timer-based interface is necessary to synchronize the model metric logger and checkpointer later.
    tasktimer = utils.SynchroTimer()  # This timer must be set manually, but it still uses model_idx as time
    task_metric_logger = utils.MetricLogger(dir_tree=dir_tree, metrics=task_metrics,
                                            set_type=utils.MetricLogger.MetricSet.task, logger=logger, timer=tasktimer)
    n_known_samples = len(task_metrics.model_idx)
    tasktimer.adjust(previous_timestamp=n_known_samples)

    task_config = {
        "train_config": train_config,
        "dataset": dataset.value[0],
        "local_seed": local_seed,
        "global_seed": global_seed,
        "debug": debug,
        "generate_sampling_profile": generate_sampling_profile,
        "nsamples": nsamples,
        "opts": opts
    }
    with open(dir_tree.task_config_file, "w") as fp:
        json.dump(task_config, fp, default=str)

    logger.setLevel(logging.DEBUG if debug else logging.INFO)
    NASB201HPOSearchSpace.NUM_CLASSES = dataset.value[3]
    search_space = NASB201HPOSearchSpace()
    if dataset.value[2] == 1:  # Check # of channels
        # The dataset is in Grayscale
        search_space.GRAYSCALE = True

    sampler = utils.model_sampler(search_space=search_space, taskid=taskid, global_seed_gen=global_seed_gen, rng=rng,
                                  portfolio_pth=portfolio_pth, opts=opts, cycle_models=cycle_portfolio)

    for model_idx, (model, model_config, curr_global_seed) in enumerate(sampler, start=1):
        if nsamples != -1 and model_idx > nsamples:
            break
        logger.info(f"Sampled new architecture: {model_config} from space {search_space.__class__.__name__}")

        ## Model initialization
        model.parse()
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        logger.debug(f"Training model {model_idx} on device {str(device)}")
        transfer_devices = device.type != "cpu"
        if transfer_devices:
            model = model.to(device)

        ## Handle task-level metrics
        if model_idx <= n_known_samples:
            # This particular model has already been sampled once. Verify config and seed.
            assert task_metrics.global_seed[model_idx - 1] == curr_global_seed, \
                f"There is a mismatch between the previously registered global seed used for evaluating the " \
                f"model index {model_idx} and the newly generated seed {curr_global_seed}"
            if not task_metrics.model_config[model_idx - 1] == model_config:
                f"Task {taskid}, model {model_idx}: Model config generation mismatch. Old model config: " \
                f"{task_metrics.model_config[model_idx - 1]}. Newly generated config: {model_config}"
            logger.debug(f"Task {taskid}, model {model_idx} has been previously sampled.")
        else:
            # A new model config has been sampled.
            task_metrics.model_idx.append(model_idx)
            task_metrics.model_config.append(model_config)
            task_metrics.global_seed.append(curr_global_seed)
            task_metrics.size_MB.append(naslib_utils.count_parameters_in_MB(model))

            tasktimer.update(timestamp=model_idx, force=True)
            task_metric_logger.log()
            logger.debug(f"Logged new sample for task {taskid}, model {model_idx}.")

        if generate_sampling_profile:
            # time.sleep(1)
            continue

        ## Actual model training and evaluation
        try:
            dir_tree.model_idx = model_idx

            if dir_tree.model_error_description_file.exists():
                os.remove(dir_tree.model_error_description_file)

            naslib_utils.set_seed(curr_global_seed)
            data_loaders, min_shape = dataset_lib.get_dataloaders(
                dataset=dataset, batch_size=train_config.batch_size, cutout=0, split=train_config.split,
                resolution=model_config.get("Resolution", 0), trivial_augment=model_config.get("TrivialAugment", False),
                datadir=datadir
            )
            validate = "valid" in data_loaders
            train(
                model=model, data_loaders=data_loaders, train_config=train_config, dir_tree=dir_tree, logger=logger,
                validate=validate, transfer_devices=transfer_devices, device=device, debug=debug, min_shape=min_shape
            )
        except Exception as e:
            naslib_logging.log_every_n_seconds(logging.INFO, "Architecture Training failed.", 15, name=logger.name)
            error_description = {
                "exception": traceback.format_exc(),
                "config": model_config,
                "global_seed": curr_global_seed
            }
            with open(dir_tree.model_error_description_file, "w") as fp:
                json.dump(error_description, fp, indent=4)
            if debug:
                raise e
        else:
            ## Clean-up after nominal program execution
            logger.info("Sampled architecture trained successfully.")

            del model  # Release memory
            dir_tree.model_idx = None  # Ensure that previously written data cannot be overwritten

    ## Clean up memory before terminating task
    del task_metrics
    del task_metric_logger
    del dir_tree


if __name__ == "__main__":
    ## Parse CLI
    args = argument_parser().parse_args()

    # Pseudo-RNG should rely on a bit-stream that is largely uncorrelated both within and across tasks
    real_taskid = args.taskid_base + args.taskid
    dataset = Datasets.__members__[args.dataset]  # Value checking has already been performed by ArgumentParser

    run_task(basedir=args.basedir, taskid=real_taskid, train_config=get_tranining_config_from_args(args),
             dataset=dataset, datadir=args.datadir, local_seed=_seed, global_seed=args.global_seed, debug=args.debug,
             generate_sampling_profile=args.generate_sampling_profile, cycle_portfolio=args.cycle_portfolio,
             nsamples=args.nsamples, portfolio_pth=args.portfolio, opts=args.opts)



"""
Test whether or not the maximum expected memory load on a particular dataset is viable for a given resource allocation.
This script launches the most expensive known architecture (all Ops are set to ConvBN3x3) for the fidelity assosciated
with each launched worker id for a specified dataset and attempts to train it for 5 epochs. This can also be used to
generate initial theoretical upper bounds on per-epoch training time for each fidelity bucket.
"""

import argparse
import logging
from pathlib import Path
import sys
from jahs_bench.tabular.lib.core import datasets as dataset_lib
from jahs_bench.tabular.lib.core.constants import Datasets
from jahs_bench.tabular.lib.core.utils import AttrDict
from jahs_bench.tabular import distributed_nas_sampling
from jahs_bench.tabular.search_space.constants import OP_NAMES

_log = logging.getLogger(__name__)

if __name__ == "__main__":

    parser = argparse.ArgumentParser("Test whether or not the maximum expected memory load on a particular dataset is "
                                     "viable for a given resource allocation.")
    parser.add_argument("--rootdir", type=Path,
                        help="The root directory within which multiple base directories will exist. See '--workerid' "
                             "for details.")
    parser.add_argument("--workerid", type=int, default=0,
                        help="A worker ID for launching this script on a cluster. When a non-negative value is given, "
                             "it is assumed that the script is operating in a cluster with multiple parallel processes "
                             "and each process will choose a sub-directory within the given 'rootdir' as its own base "
                             "directory based on the fidelity parameter values. Raises an error when a negative value "
                             "is given.")
    parser.add_argument("--dataset", type=str, choices=list(Datasets.__members__.keys()),
                        help="The name of which dataset is to be used for model training and evaluation. Only one of "
                             "the provided choices can be used.")
    parser.add_argument("--datadir", type=Path, default=dataset_lib.get_default_datadir(),
                        help="The directory where all datasets are expected to be stored.")
    parser.add_argument("--debug", action="store_true", help="When given, enables debug level output.")
    args = parser.parse_args()

    if args.debug:
        _log.setLevel(logging.DEBUG)
    else:
        _log.setLevel(logging.INFO)

    wid = args.workerid

    if wid < 0:
        raise ValueError(f"Worker id must be a non-negative integer, was {args.workerid}.")

    fmt = logging.Formatter(f"*Worker {wid}* [%(asctime)s] %(name)s %(levelname)s: "
                            f"%(message)s", datefmt="%m/%d %H:%M:%S")
    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(fmt)
    _log.addHandler(ch)

    fids = {"N": [1, 3, 5], "W": [4, 8, 16], "Resolution": [0.25, 0.5, 1.0]}
    if wid >= 27:
        sys.exit(0)
    n, w, r = fids["N"][wid % 3], fids["W"][(wid // 3) % 3], fids["Resolution"][(wid // 9) % 3]
    ids = [n, w, r]
    subdir = "-".join([f"{k}-{ids[i]}" for i, k in enumerate(fids.keys())])

    basedir: Path = args.rootdir / subdir / "tasks"
    assert args.rootdir.exists() and args.rootdir.is_dir()
    basedir.mkdir(parents=True, exist_ok=True)

    dataset = Datasets.__members__[args.dataset]
    train_config = AttrDict(epochs=5, batch_size=256, use_grad_clipping=False, split=True, warmup_epochs=0,
                            disable_checkpointing=False, checkpoint_interval_seconds=None, checkpoint_interval_epochs=1)

    op_name = "ConvBN3x3"
    op_index = OP_NAMES.index(op_name)
    base_opts = " ".join([" ".join([k, str(ids[i])]) for i, k in enumerate(fids.keys())] + ["Optimizer SGD"])
    arch_opts = " ".join([f"Op{i} {op_index}" for i in range(1, 7)])  # Set all edge operations
    opts = " ".join([base_opts, arch_opts]).split(" ")

    _log.info(f"Worker {wid}: Beginning memory usage check at {basedir} for dataset {dataset.value[0]}.")

    try:
        distributed_nas_sampling.run_task(
            basedir=basedir, taskid=args.workerid, train_config=train_config, dataset=dataset, datadir=args.datadir,
            local_seed=distributed_nas_sampling._seed, global_seed=None, debug=False, generate_sampling_profile=False,
            nsamples=1, portfolio_pth=None, cycle_portfolio=False, logger=_log, opts=opts
        )
    except Exception as e:
        _log.warning(f"Worker {wid}: Ran into an error while running the model at {basedir}. Exiting. Cause: {str(e)}")
        sys.exit(1)

    _log.info(f"Worker {wid}: Finished memory usage test.")



"""
A small script to extract the best performing models from all available configurations
generated after an HPO run using the NEPS framework.
"""
import argparse
import functools
import logging
from pathlib import Path
import shutil
import yaml
from typing import Sequence, Iterator, Dict, Tuple, Optional

_log = logging.getLogger(__name__)

def get_loss(config_dir: Path) -> float:
    """ Given the path to a particular configuration's working directory, extract the
    configuration's recorded loss and return it. """

    result_file = config_dir / "result.yaml"
    with open(result_file) as fp:
        result = yaml.safe_load(fp)

    return result["loss"]

def iterate_configs(metric_dir: Path, max_configs: Optional[int] = None) -> Iterator[Path]:
    """ Given the working directory corresponding to a single metric, iterate over all
    the individual configurations' working directories. """

    subdir = metric_dir / "results"
    config_dirs = subdir.iterdir() if max_configs is None \
        else [subdir / f"config_{i}" for i in range(1, max_configs + 1)]
    for config_dir in config_dirs:
        yield config_dir

def iterate_metrics(root_dir: Path, common_suffix_dirs: Optional[Sequence[str]] = None) \
        -> Iterator[Tuple[str, Path]]:
    """ Given a root directory within which, after descending down an optional sequence
    of sub-directories common to each directory tree, the working directories for each
    metric can be found, return a generator to iterate over these working directories. """

    for subdir in root_dir.iterdir():
        metric_dir = subdir
        metric_name = subdir.name
        if common_suffix_dirs is not None:
            metric_dir = functools.reduce(lambda a, x: a / x, common_suffix_dirs,
                                          metric_dir)
        yield metric_name, metric_dir

def identify_best_configs(root_dir: Path, common_suffix_dirs: Optional[
                          Sequence[str]] = None, max_configs: Optional[int] = None) \
        -> Dict[str, Tuple[Path, float]]:
    """ Given a root directory containing all the metric directories for a single
    dataset's surrogates (after optionally further descending down a common set of
    sub-directories), returns a dictionary mapping each metric to the working directory
    of the best performing model configuration and its loss value. When 'max_configs' is
    given, only the first 'max_configs' configurations are considered. """

    best_configs = {}
    for metric, metric_dir in iterate_metrics(root_dir, common_suffix_dirs):
        best_config = None
        min_loss = None
        _log.info(f"Identifying best configuration for metric {metric}.")
        for config_dir in iterate_configs(metric_dir, max_configs):
            loss = get_loss(config_dir)
            if best_config is None:
                min_loss = loss
                best_config = config_dir
            elif loss < min_loss:
                min_loss = loss
                best_config = config_dir
        best_configs[metric] = (best_config, min_loss)
        _log.info(f"Best configuration for metric {metric} found at {best_config} with "
                  f"loss {min_loss}.")

    return best_configs

def assemble_surrogate(best_configs: Dict[str, Tuple[Path, float]], final_dir: Path) \
        -> Path:
    """ Given a mapping from metric names to a tuple containing the path to the best
    configuration and its loss for that metric, copies all the models into the new
    directory tree that can be used by the JAHS-Bench-201 public API to load a surrogate
    model. The path to this directory is returned. """

    final_dir.mkdir(exist_ok=True, parents=False)
    for metric, (config_dir, loss) in best_configs.items():
        model_dir = config_dir / "xgb_model"
        pth: Path = shutil.copytree(model_dir, final_dir / metric)
        _log.info(f"Inserted best configuration for metric '{metric}' at {pth}")

    return final_dir

def main(final_dir: Path, root_dir: Path, max_configs: Optional[int] = None,
         common_suffix_dirs: Optional[Sequence[str]] = None):
    """ Executes the main program control. """

    assert root_dir.exists() and root_dir.is_dir()
    _log.info(f"Identifying the best configuration for each metric from those present "
              f"in {root_dir}.")
    best_configs = identify_best_configs(root_dir, common_suffix_dirs, max_configs)

    _log.info(f"Assembling the individual models into one directory at {final_dir}.")
    final_model_dir = assemble_surrogate(best_configs, final_dir)

    _log.info(f"Assembled the final model at: {final_model_dir}")

def parse_cli():
    parser = argparse.ArgumentParser(
        "Assemble the best configurations found for each metric after HPO using NEPS "
        "into a single ensemble of surrogate models usable by the public API. "
    )
    parser.add_argument("--final_dir", type=Path,
                        help="The final directory where the assembled ensemble should be "
                             "stored.")
    parser.add_argument("--root_dir", type=Path,
                        help="Path to the working directory where, for a single task, "
                             "all the metrics' evaluated configurations are stored.")
    parser.add_argument("--max_configs", type=int, default=None,
                        help="Maximum number of configs to consider when choosing the "
                             "best performing config. When given, only the first "
                             "'max-configs' configs are considered, otherwise all "
                             "available configs are considered.")
    parser.add_argument("--common_suffix_dirs", type=str, default=None,
                        nargs=argparse.REMAINDER,
                        help="An optional sequence of strings defining further "
                             "sub-directories beyond 'root_dir/<metric_name/' to be "
                             "traversed for each metric before the directory containing "
                             "configuration specific working directories is reached.")

    args = parser.parse_args()
    return args

if __name__ == "__main__":
    args = parse_cli()
    logging.basicConfig(level=logging.INFO,
                        format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
                        datefmt="%m/%d %H:%M:%S")
    _log.setLevel(logging.INFO)
    main(**vars(args))



from __future__ import annotations

import copy
import logging
from functools import partial
from pathlib import Path
from typing import Union, Optional, Sequence, Tuple, Dict

import ConfigSpace
import joblib
import numpy as np
import pandas as pd
import scipy.stats
import sklearn
import sklearn.compose
import sklearn.metrics
import sklearn.model_selection
import sklearn.multioutput
import sklearn.pipeline
import sklearn.preprocessing
import xgboost as xgb
import yacs.config as config
from jahs_bench.lib.core import utils as core_utils
from jahs_bench.lib.core.configspace import joint_config_space
from jahs_bench.surrogate import utils as surrogate_utils, config

_log = logging.getLogger(__name__)
ConfigType = Union[dict, ConfigSpace.Configuration]


class XGBSurrogate:
    """ A surrogate model based on XGBoost. """

    estimators_per_output: int
    hyperparams: dict
    config_space: ConfigSpace.ConfigurationSpace
    label_headers: Optional[pd.Index]
    feature_headers: Optional[pd.Series]
    trained_: bool

    __params_filename = "params.pkl.gz"
    __headers_filename = "label_headers.pkl.gz"
    __model_filename = "model.pkl.gz"
    __param_keys = ["estimators_per_output", "hyperparams", "config_space",
                    "label_headers", "feature_headers",
                    "trained_"]
    __objective = "reg:squarederror"

    _hpo_search_space = {
        "objective": [__objective],
        # "eval_metric": "rmse",
        # 'early_stopping_rounds': 100,
        "booster": ["gbtree"],
        "max_depth": np.arange(1, 15),
        "min_child_weight": list(range(1, 10)),
        "colsample_bytree": scipy.stats.uniform(0.0, 1.0),
        "learning_rate": scipy.stats.loguniform(0.001, 0.5),
        # 'alpha': 0.24167936088332426,
        # 'lambda': 31.393252465064943,
        "colsample_bylevel": scipy.stats.uniform(0.0, 1.0),
    }

    @property
    def default_hyperparams(self) -> dict:
        params = {
            "objective": self.__objective,
            # "eval_metric": "rmse",
            "booster": "gbtree",
            "max_depth": 6,
            "min_child_weight": 1,
            "colsample_bytree": 1,
            "learning_rate": 0.3,
            "colsample_bylevel": 1,
        }
        return params

    def set_random_hyperparams(self):
        if self.hyperparams is None:
            # evaluate the default config first during HPO
            params = self.default_hyperparams.copy()
        else:
            params = {
                # "objective": "reg:squarederror",
                "objective": self.__objective,
                # "eval_metric": "rmse",
                # 'early_stopping_rounds': 100,
                "booster": "gbtree",
                "max_depth": int(np.random.choice(range(1, 15))),
                "min_child_weight": int(np.random.choice(range(1, 10))),
                "colsample_bytree": np.random.uniform(0.0, 1.0),
                "learning_rate": loguniform(0.001, 0.5),
                # 'alpha': 0.24167936088332426,
                # 'lambda': 31.393252465064943,
                "colsample_bylevel": np.random.uniform(0.0, 1.0),
            }
        self.hyperparams = params
        return params

    def __init__(self, config_space: Optional[
        ConfigSpace.ConfigurationSpace] = joint_config_space,
                 estimators_per_output: int = 500, use_gpu: Optional[bool] = None,
                 hyperparams: dict = None):
        """
        Initialize the internal parameters needed for the surrogate to understand the
        data it is dealing with.
        :param config_space: ConfigSpace.ConfigurationSpace
            A config space to describe what each model config looks like.
        :param estimators_per_output: int
            The number of trees that each XGB forest should boost. Experimental, will
            probably be removed in favour of a dynamic approach soon.
        :param use_gpu: bool
            A flag to ensure that a GPU is used for model training. If False (default),
            the decision is left up to XGBoost itself, which in turn depends on being
            able to detect a GPU.
        """

        self.config_space = config_space
        self.estimators_per_output = estimators_per_output
        self.hyperparams = hyperparams
        self.use_gpu = use_gpu
        self.model = None
        self.feature_headers = None
        self.label_headers = None
        self.trained_ = False

        # Both initializes some internal attributes as well as performs a sanity test
        if self.hyperparams is None:
            self.set_random_hyperparams()  # Sets default hyperparameters

    @property
    def preprocessing_pipeline(self):
        """ The pre-defined pre-processing pipeline used by the surrogate. """

        params = self.config_space.get_hyperparameters()

        # TODO: Add cache dir to speed up HPO
        # TODO: Read categorical choices from the config space
        onehot_columns = [p.name for p in params if
                          isinstance(p, ConfigSpace.CategoricalHyperparameter)]
        onehot_enc = sklearn.preprocessing.OneHotEncoder(drop="if_binary")
        transformers = [("OneHotEncoder", onehot_enc, onehot_columns)]
        prep_pipe = sklearn.compose.ColumnTransformer(transformers=transformers,
                                                      remainder="passthrough")
        return prep_pipe

    # TODO: Also handle assigning appropriate dtypes in Pandas DataFrame
    def _random_data(
        self, nconfigs: int = 10, samples_per_config: int = 100, nlabel_dims: int = 2,
        label_names: Optional[Sequence[str]] = None,
        random_state: Optional[np.random.RandomState] = None,
        config_space_constraints: Optional[Dict[str, Any]] = None) -> \
            Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """ A debugging tool. Generate a random dataset of arbitrary size using the
        stored config space representation. Returns the randomly generated set of
        features, labels and groups as pandas DataFrame objects, in that order. A
        dictionary mapping the respective config space parameter names to appropriate
        values may be passed in 'config_space_consts' in order to restrict the sampling
        such that a subset of parameters always have the specified values. Note that this
        affects the reproducibility as well as statistical independence of the sampling
        procedure, in that these properties are only guaranteed for the modified search
        space. The resultant configurations will still be compatible with the original
        search space and any assosciated pre-processing pipelines, as long as the
        specified constant values are also compatible. """

        cs = copy.deepcopy(self.config_space)
        if config_space_constraints is not None:
            cs, _ = core_utils.adapt_search_space(original_space=cs,
                                                  opts=config_space_constraints)

        if not isinstance(random_state, np.random.RandomState):
            # Assume that rng is either None or a compatible source of entropy
            random_state = np.random.RandomState(random_state)

        cs.random = random_state

        features = cs.sample_configuration(nconfigs)
        features = np.array([list(c.get_dictionary().values()) for c in features])
        features = np.repeat(features, samples_per_config, axis=0)
        features = pd.DataFrame(features, columns=cs.get_hyperparameter_names())
        epochs = pd.Series(np.tile(np.arange(1, samples_per_config + 1), nconfigs),
                           name="epoch")
        features.loc[:, "epoch"] = epochs

        labels = random_state.random((nconfigs * samples_per_config, nlabel_dims))
        label_names = [f"Label_{i}" for i in
                       range(nlabel_dims)] if label_names is None else label_names
        labels = pd.DataFrame(labels, columns=label_names)

        groups = np.repeat(np.arange(1, nconfigs + 1), samples_per_config, axis=0)
        groups = pd.DataFrame(groups, columns=["ModelIndex"])

        return features, labels, groups

    def _get_simple_pipeline(self, multiout: bool = True) -> sklearn.pipeline.Pipeline:
        """
        Get a Pipeline instance that can be used to train a new surrogat model. This is
        the simplest available pipeline that simply normalizes all outputs and fits an
        XGBoost regressor to each regressand to be predicted. HPO is performed by simple
        random search over a single set of hyperparameters common to all regressors.

        :return: pipeline
        """

        prep_pipe = self.preprocessing_pipeline
        xgboost_estimator = xgb.sklearn.XGBRegressor(
            n_estimators=500, tree_method="gpu_hist" if self.use_gpu else "auto",
            n_jobs=1, **self.hyperparams)

        if multiout:
            multi_regressor = sklearn.multioutput.MultiOutputRegressor(
                estimator=xgboost_estimator, n_jobs=1)
            pipeline_steps = [
                ("preprocess", prep_pipe),
                ("multiout", multi_regressor)
            ]
        else:
            pipeline_steps = [
                ("preprocess", prep_pipe),
                ("estimator", xgboost_estimator)
            ]

        pipeline = sklearn.pipeline.Pipeline(steps=pipeline_steps)
        return pipeline

    def _build_pipeline(self, pipeline_config: config.CfgNode):
        """ Construct a customized pipeline using various parameters. """

        _log.debug(f"Constructing a custom pipeline using the given pipeline "
                   f"configuration.")

        # Construct the objective function
        loss_type = config.RegressionLossFuncTypes(pipeline_config.loss)
        _log.debug(f"Using the '{loss_type.name}' objective.")
        if loss_type is config.RegressionLossFuncTypes.custom:
            self.__objective = \
                surrogate_utils.custom_loss_function(pipeline_config.loss_params)
        else:
            self.__objective = loss_type.value
        self.hyperparams["objective"] = self.__objective

        # Build input preprocessing pipeline and bare-bones XGBoost estimator
        prep_pipe = self.preprocessing_pipeline
        xgboost_estimator = xgb.sklearn.XGBRegressor(
            n_estimators=500, tree_method="gpu_hist" if self.use_gpu else "auto",
            n_jobs=1, **self.hyperparams)

        # Target processing pipeline steps
        target_config = pipeline_config.target_config
        assert target_config.transform is None or isinstance(target_config.transform, str)

        if not target_config.transform:
            estimator = xgboost_estimator
        else:
            estimator = surrogate_utils.apply_target_transform(
                target_config=target_config, final_estimator=xgboost_estimator)

        # Final pipeline generation - multiout will be handled by the public API
        pipeline_steps = [
            ("preprocess", prep_pipe),
            ("estimator", estimator)
        ]

        pipeline = sklearn.pipeline.Pipeline(steps=pipeline_steps)
        _log.debug(f"Finished generating new pipeline.")
        return pipeline

    @staticmethod
    def prepare_dataset_for_training(
        features: pd.DataFrame, labels: pd.DataFrame,
        groups: Optional[pd.DataFrame] = None, test_size: float = 0.,
        random_state: Optional[np.random.RandomState] = None, num_cv_splits: int = 5,
        stratify: bool = True, strata: Optional[pd.Series] = None
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame,
               Optional[pd.DataFrame], Optional[pd.DataFrame],
               sklearn.model_selection.BaseCrossValidator]:
        """
        Given a full dataset consisting of the input features, the labels to be
        predicted, and an optional definition of groups for the data, ascertains the way
        the data should be split into a training, validation and test set to be used for
        training a surrogate pipeline.

        :param features: pandas.DataFrame
        :param labels: pandas.DataFrame
        :param groups: pandas.DataFrame or None
            A set of additional numeric labels for each row in 'features' and 'labels'
            such that all rows with the same label are assosciated with the same model
            config and as such will be treated as one unit during all dataset shuffling
            and splitting operations. Can be omitted (i.e. set to None)
        :param test_size: float
            A real number in the range [0., 1.) to indicate what fraction of the given
            data should be split off into a test dataset. A value of 0. (default)
            indicates that no test set should be generated.
        :param random_state:
        :param num_cv_splits: int
            An integer >=2, the number of cross-validation splits to be generated from
            the training set. Can be an arbitrary number that won't affect model
            training if no Cross-Validation is actually performed during model training.
        :param stratify: bool
            When True, output strata distribution is also considered when splitting the
            data. Also consult ´strata´.
        :param strata: None or string or pandas Series
            The strata according to which splitting occurs, such that a best effort is
            made to maintain the distribution of either the groups (if given) or the
            raw data rows across the given strata. If None, and ´stratify=True´, the
            first column in the labels is used as the strata. A string can also be passed
            to indicate which column from labels DataFrame is to be used as strata.
            Otherwise, this should be a pandas Series or other iterable with a length
            corresponding to the number of rows in the dataset. No effect when
            ´stratify=False´.
        :return:
            A tuple containing the training features, test features, training labels,
            test labels, an optional set of corresponding groups for the training data
            and the cross-validation split generator, in that order. The cross-validation
            split generator can be used to split the training data into training and
            validation splits on the fly.
        """

        _log.info(
            "Generating training and test splits, and the validation split generator.")
        if test_size > 1. or test_size < 0.:
            raise ValueError(
                f"The test set fraction size 'test_size' must be in the range [0, 1), "
                f"was given {test_size}")
        elif test_size == 0.:
            _log.debug("No test split will be generated.")
            xtrain, ytrain = features, labels
            xtest, ytest = None, None
            if groups is None:
                _log.debug("No data groups were given.")
                splitter_type = sklearn.model_selection.KFold
            else:
                _log.debug("Generating training and validation splits in accordance with "
                           "the given data groups.")
                splitter_type = sklearn.model_selection.GroupKFold
        else:
            _log.debug(f"Generating a test split containing approximately "
                       f"{test_size * 100} % of the total data.")
            strata = labels.loc[:, labels.columns[0]] if strata is None else \
                labels.loc[:, strata] if isinstance(strata, str) else strata

            if groups is None:
                _log.debug("No data groups were given.")
                splitter_type = sklearn.model_selection.StratifiedShuffleSplit \
                    if stratify else sklearn.model_selection.ShuffleSplit
                test_splitter = splitter_type(n_splits=1, test_size=test_size,
                                              random_state=random_state)

                idx_train, idx_test = next(test_splitter.split(features, strata))
            else:
                _log.debug(
                    "Generating training, validation and test splits in accordance with "
                    "the given data groups.")
                splitter_type = partial(sklearn.model_selection.StratifiedGroupKFold,
                                        shuffle=True) \
                    if stratify else sklearn.model_selection.GroupShuffleSplit
                test_splitter = splitter_type(n_splits=int(1 / test_size),
                                              random_state=random_state)

                idx_train, idx_test = next(
                    test_splitter.split(features, strata, groups=groups))
                groups = groups.iloc[idx_train]

            xtrain = features.iloc[idx_train]
            xtest = features.iloc[idx_test]
            ytrain = labels.iloc[idx_train]
            ytest = labels.iloc[idx_test]
            strata = strata.iloc[idx_train] if stratify else strata

        cv = splitter_type(n_splits=num_cv_splits, random_state=random_state)

        _log.info("Dataset splits successfully generated.")
        return xtrain, xtest, ytrain, ytest, groups, strata, cv

    # TODO: Check if a train split, after test split has been taken out, would have
    #  sufficient representation in terms of categorical values (at least one occurence
    #  of each)
    def fit(self, features: pd.DataFrame, labels: pd.DataFrame,
            random_state: np.random.RandomState = None,
            pipeline_config: Optional[config.CfgNode] = None):
        """
        Pre-process the given dataset, fit an XGBoost model on it and return the training
        errors.

        :param features: DataFrame
            The inputs to the predictive model.
        :param labels: DataFrame
            The known labels the model is expected to predict for the given inputs.
        :param random_state: NumPy RandomState object or None
            A random number generator. If it's None, local system entropy is used to
            generate an RNG on the spot. The latter will break reproducibility.
        :param pipeline_config: ConfigNode or None
            A configuration object specifying how to build the training pipeline. Consult
            `XGBSurrogate._build_pipeline()` and `surrogate.constants` for more details.
        :return:
        """

        # Ensure the order of the features does not get messed up and is always accessible
        if self.feature_headers is None:
            self.feature_headers = features.columns
        else:
            features = features.loc[:, self.feature_headers]

        # Ensure the order of labels does not get messed up and is always accessible
        if self.label_headers is None:
            self.label_headers = labels.columns
        else:
            labels = labels.loc[:, self.label_headers]

        # TODO: Implement early stopping to determine correct final value for
        #  n_estimators - NASLib used a fixed value of 500, so this procedure may or may
        #  not be useful and certainly needs a reference. This is a method to prevent
        #  overfitting, analogous to cutting off NN training after a certain number of
        #  epochs.

        num_regressands = labels.columns.size
        pipeline = self._get_simple_pipeline(multiout=num_regressands > 1) \
            if not pipeline_config else self._build_pipeline(pipeline_config)

        self.model = pipeline.fit(features, labels)
        self.trained_ = True

        ypred_train = self.predict(features)
        train_r2 = sklearn.metrics.r2_score(labels, ypred_train)
        train_mse = sklearn.metrics.mean_squared_error(labels, ypred_train)
        scores = {
            "train_r2": train_r2,
            "train_mse": train_mse,
        }

        return scores

    def predict(self, features: pd.DataFrame) -> pd.DataFrame:
        """ Given some input data, generate model predictions. The input data will be
        properly encoded when this function is called. """

        features = features.loc[:, self.feature_headers]
        ypredict = self.model.predict(features)
        ypredict = pd.DataFrame(ypredict, columns=self.label_headers)
        return ypredict

    def dump(self, outdir: Path, protocol: int = 0):
        """ Save a trained surrogate to disk so that it can be loaded up later. """

        params = {k: self.__getattribute__(k) for k in self.__param_keys}
        joblib.dump(params, outdir / self.__params_filename, protocol=protocol)
        if self.trained_:
            self.label_headers.to_series().to_pickle(outdir / self.__headers_filename,
                                                     protocol=protocol)
            joblib.dump(self.model, outdir / self.__model_filename, protocol=protocol)

    @classmethod
    def load(cls, outdir: Path) -> XGBSurrogate:
        """ Load a previously saved surrogate from disk and return it. """

        params: dict = joblib.load(outdir / cls.__params_filename)
        surrogate = cls()
        for k, v in params.items():
            surrogate.__setattr__(k, v)

        if surrogate.trained_:
            label_headers: pd.Series = pd.read_pickle(outdir / cls.__headers_filename)
            model = joblib.load(outdir / cls.__model_filename)

            surrogate.label_headers = pd.Index(label_headers)
            surrogate.model = model

        return surrogate



"""
This is mostly a set of convenience wrappers around small sequences of operations on collated metric DataFrames that
make it easier to post-process the data in a semantic manner. Another advantage of using the convenience functions here
is that they handle a lot of boiler-plate code for implementing various checks and balances and ensuring consistent
outputs.

Expected structure of a single job's metric DataFrame:

index - pandas.MultiIndex, levels: [taskid, model_idx, Epoch]
columns - pandas.MultiIndex, level names: [MetricType, MetricName]
    where MetricType is generally [train, valid, test, diagnostic] and MetricName corresponds to values from
    tabular_sampling.lib.constants

"""

import argparse
import logging
from functools import wraps
from pathlib import Path
from typing import Dict, Iterable, Union, Optional, Any

import pandas as pd

from jahs_bench.tabular.lib.core.constants import MetricDFIndexLevels
from jahs_bench.tabular.search_space.constants import OP_NAMES, EDGE_LIST

_log = logging.getLogger(__name__)

model_ids_by = [MetricDFIndexLevels.taskid.value, MetricDFIndexLevels.modelid.value]


# fidelity_params=("N", "W")

# Common settings
# plt.rcParams.update({"xtick.labelsize": 14, "ytick.labelsize": 14, "axes.labelsize": 16, "axes.titlesize": 16})


def map_label_to_color(label: str, map: Dict, colors: Iterable):
    if not label in map:
        map[label] = next(colors)
    return map[label]


def _parse_cli():
    parser = argparse.ArgumentParser()
    parser.add_argument("--basedir", type=Path, help="Path to the base directory where the relevant dataframe(s) "
                                                     "is/are stored as a file named data.pkl.gz.")
    args = parser.parse_args()
    return args


def load_metric_df(basedir: Path = None, df: pd.DataFrame = None) -> pd.DataFrame:
    assert basedir is not None or df is not None, "Either the 'basedir' should be a Path to load the data from " \
                                                  "or 'df' should be a pre-loaded metric data DataFrame."
    if df is None:
        # TODO: Make the metric df pickle file name a reusable constant
        df_fn = basedir / "metrics.pkl.gz"
        df: pd.DataFrame = pd.read_pickle(df_fn)
    else:
        assert isinstance(df, pd.DataFrame), f"This script was designed to only work with pandas DataFrames, not " \
                                             f"{type(df)}"

    return df


def _df_loader_wrapper(func):
    @wraps(func)
    def wrapper(basedir: Path = None, df: pd.DataFrame = None, *args, **kwargs) -> pd.DataFrame:
        df = load_metric_df(basedir=basedir, df=df)
        return func(basedir, df, *args, **kwargs)

    return wrapper


@_df_loader_wrapper
def get_runtimes(basedir: Path, df: pd.DataFrame, display: bool = False, reduce_epochs: bool = True,
                 extra_durations: list = None) -> pd.DataFrame:
    """
    Extract the runtimes of the various models.

    :param basedir:
    :param df:
    :param display: bool
        When True, logs the runtime statistics.
    :param reduce_epochs: bool
        When True, abstracts away the per-epoch metrics and generates runtimes on a per-model basis instead.
    :return:
    """

    datasets = ["train", "valid", "test"]
    duration_metrics = ["duration"] + (extra_durations if extra_durations is not None else [])
    runtime_metrics = [(ds, met) for ds in datasets for met in duration_metrics] + [("diagnostic", "runtime")]

    if reduce_epochs:
        d1 = df.loc[:, runtime_metrics[:-1]].groupby(model_ids_by).agg("sum")
        d2 = df.loc[:, runtime_metrics[-1:]].groupby(model_ids_by).agg("max")
        d1.columns = d1.columns.map("{0[0]}-{0[1]}".format)
        d2.columns = d2.columns.map("{0[1]}".format)
        runtimes_df = d1.join([d2])
    else:
        runtimes_df = df[runtime_metrics]

    if display:
        _log.info(f"Runtime statistics:\n{runtimes_df.describe()}")
    return runtimes_df


@_df_loader_wrapper
def get_nepochs(basedir: Path, df: pd.DataFrame, filter_epochs: int = -1) -> pd.DataFrame:
    assert isinstance(df.index, pd.MultiIndex), f"The input DataFrame index must be a MultiIndex, was {type(df.index)}"
    assert MetricDFIndexLevels.epoch.value in df.index.names, \
        f"Missing the level '{MetricDFIndexLevels.epoch.value}' for epochs in input DataFrame with MultiIndex index " \
        f"levels {df.index.names}."

    nepochs = df[df.columns.values[0]]
    nepochs = nepochs.groupby(model_ids_by).agg("count").to_frame(("nepochs"))

    if filter_epochs > 0:
        nepochs = nepochs[nepochs["nepochs"].where(nepochs["nepochs"] == filter_epochs).notna()]

    return nepochs


@_df_loader_wrapper
def get_configs(basedir: Path, df: pd.DataFrame) -> pd.DataFrame:
    assert isinstance(df.index, pd.MultiIndex), f"The input DataFrame index must be a MultiIndex, was {type(df.index)}"
    assert MetricDFIndexLevels.epoch.value in df.index.names, \
        f"Missing the level '{MetricDFIndexLevels.epoch.value}' for epochs in input DataFrame with MultiIndex index " \
        f"levels {df.index.names}."
    assert all([l in df.index.names for l in model_ids_by]), \
        f"The input DataFrame index must include the levels {model_ids_by}, but had the levels {df.index.names}."

    confs = df["model_config"].xs(1, level=MetricDFIndexLevels.epoch.value)
    confs = confs.reorder_levels(model_ids_by, axis=0)

    return confs


@_df_loader_wrapper
def get_accuracies(basedir: Path, df: pd.DataFrame, include_validation: bool = False) -> pd.DataFrame:
    """
    Extract and return a DataFrame containing only the train, test and (if enabled) validation accuracy scores.
    """

    # outdir = basedir / "analysis"
    # outdir.mkdir(exist_ok=True, parents=True)

    assert isinstance(df.index, pd.MultiIndex), f"The input DataFrame index must be a MultiIndex, was {type(df.index)}"
    assert all([l in df.index.names for l in model_ids_by]), \
        f"The input DataFrame index must include the levels {model_ids_by}, but had the levels {df.index.names}."

    train_acc: pd.DataFrame = df[("train", "acc")].groupby(model_ids_by).agg("max").to_frame("train-acc")
    test_acc: pd.DataFrame = df[("test", "acc")].groupby(model_ids_by).agg("max").to_frame("test-acc")
    if include_validation:
        valid_acc: pd.DataFrame = df[("valid", "acc")].groupby(model_ids_by).agg("max").to_frame("valid-acc")
        acc_df = test_acc.join([train_acc, valid_acc])
    else:
        acc_df = test_acc.join([train_acc])

    return acc_df


@_df_loader_wrapper
def get_losses(basedir: Path, df: pd.DataFrame, include_validation: bool = False) -> pd.DataFrame:
    """
    Extract and return a DataFrame containing only the train, test and (if enabled) validation losses.
    """

    # outdir = basedir / "analysis"
    # outdir.mkdir(exist_ok=True, parents=True)

    assert isinstance(df.index, pd.MultiIndex), f"The input DataFrame index must be a MultiIndex, was {type(df.index)}"
    assert all([l in df.index.names for l in model_ids_by]), \
        f"The input DataFrame index must include the levels {model_ids_by}, but had the levels {df.index.names}."

    train_loss: pd.DataFrame = df[("train", "loss")].groupby(model_ids_by).agg("max").to_frame("train-loss")
    test_loss: pd.DataFrame = df[("test", "loss")].groupby(model_ids_by).agg("max").to_frame("test-loss")
    if include_validation:
        valid_loss: pd.DataFrame = df[("valid", "loss")].groupby(model_ids_by).agg("max").to_frame("valid-loss")
        loss_df = test_loss.join([train_loss, valid_loss])
    else:
        loss_df = test_loss.join([train_loss])

    return loss_df

    # if "idx" in df.index.names:
    #     df_no_idx = df.xs(0, level="idx")
    # else:
    #     df_no_idx = df
    #
    # fidelity_group = df_no_idx.groupby(level=fidelity_params)
    # counts = fidelity_group.size()
    # if save_tables:
    #     table = counts.unstack(level=fidelity_params[-1]).to_latex(caption="Fidelity-wise sample count.")
    #     with open(outdir / "fidelity_count_table.tex", "w") as fp:
    #         fp.write(table)
    # else:
    #     print(f"Fidelity-wise sample count:\n{counts.unstack(level=fidelity_params[-1])}")
    #
    # fig, ax = plt.subplots(1, 1, figsize=(16, 9)) if save_plots else plt.subplots(1, 1)
    # fig: plt.Figure
    # ax: plt.Axes
    #
    # bar_width = 1
    # bin_loc_offset = 1
    # xticks = [[], []]
    # labels_to_colors = {}
    # all_colors = iter(plt.cm.Set2.colors)
    #
    # for level_0_key in counts.index.unique(0):
    #     subseries = counts.xs(level_0_key, level=0)
    #     ser_colors = [map_label_to_color(i, labels_to_colors, all_colors) for i in subseries.keys()]
    #     bin_locs = list(range(bin_loc_offset, bin_loc_offset + subseries.size, bar_width))
    #     ax.bar(bin_locs, height=subseries.values, width=bar_width, color=ser_colors)
    #     bin_loc_offset += subseries.size + 2
    #     xticks[0] += [sum(bin_locs) / subseries.size, ]
    #     xticks[1] += [level_0_key, ]
    #
    # ax.xaxis.set_ticks(xticks[0])
    # ax.xaxis.set_ticklabels(xticks[1])
    #
    # legend_labels = [f"{counts.index.names[1]}={k}" for k in labels_to_colors.keys()]
    # legend_lines = [Line2D([0], [0], linewidth=8, color=labels_to_colors[l]) for l in labels_to_colors.keys()]
    # ax.legend(legend_lines, legend_labels, loc="upper right", fontsize=16)
    # ax.set_title("Number of samples drawn from each fidelity level.")
    # ax.set_xlabel("Number of cell-repetitions per architecture (N).")
    # ax.set_ylabel("Number of samples drawn from the search space.")
    # ax.yaxis.set_tick_params(which='minor')
    # # ax.yaxis.grid(True, which='major')
    # # ax.yaxis.grid(True, which='minor', linestyle="--", linewidth=2., color='gray')
    # if save_plots:
    #     fig.tight_layout()
    #     fig.savefig(outdir / "per_fidelity_sample_counts.pdf")
    # else:
    #     plt.show()
    #
    # ### Analyze average model training time ###
    # mean_train_times = df["train_time"].groupby(level=df.index.names.difference(["idx"])).mean()
    # mean_train_times: pd.Series = mean_train_times.groupby(level=fidelity_params).mean()
    #
    # if save_tables:
    #     table = mean_train_times.to_frame().unstack(fidelity_params[-1]).to_latex(
    #         caption="Average training time per epoch across fidelity values.")
    #     with open(outdir / "per_fidelity_training_time.tex") as fp:
    #         fp.write(table)
    # else:
    #     print(f"Fidelity-wise average model training time per epoch:\n"
    #           f"{mean_train_times.to_frame().unstack(fidelity_params[-1])}")

    # fig, ax = plt.subplots(1, 1, figsize=(16, 9)) if save_plots else plt.subplots(1, 1)
    # fig: plt.Figure
    # ax: plt.Axes


# @df_loader_wrapper
# def group(basedir: Path = None, df: pd.DataFrame = None, groupby: list = None, columns: list = None):
#     """ Convenience function for generating specific groupings. The list "groupby" refers to column names that should
#     be used to group the data by and the list "columns" refers to those columns that should be present in the grouped
#     data. If "groupby" is None, the dataframe is returned unchanged. If "columns" is None, all columns other than those
#     in "groupby" are present in the returned group. """
#
#     if groupby is None:
#         return df, df
#
#     assert all([g in df.columns.names for g in groupby]), \
#         f"Mismatch between provided grouping parameters, {groupby}, and the dataframe's column names {df.columns.names}"
#
#     g = df.groupby(groupby)
#
#     pass


def collapse_index_names(index: pd.MultiIndex, levels: list = None, nlevels: int = None, separator: str = "-") -> \
        Union[pd.Index, pd.MultiIndex]:
    """ Collapses the index labels in a MultiIndex. The levels to be collapsed can be passed as either a list of
    specific levels in "levels" or an integer can be passed to "nlevels" indicating that the first "nlevels" levels
    should be collapsed into one level - the first level. """

    if levels is not None and nlevels is not None:
        raise RuntimeError("Either a list of index names or the number of indices to be collapsed should be provided, "
                           "not both.")

    if levels is not None:
        inds = [index.names.index(i) for i in levels]
        old_names = levels
    else:
        inds = list(range(nlevels))
        old_names = [index.names[i] for i in inds]

    compressor = separator.join([f"{{0[{i}]}}" for i in inds]).format

    if len(inds) > index.nlevels:
        raise RuntimeError(f"Number of levels to be collapsed - {len(inds)} - cannot exceed the number of levels in "
                           f"the original index - {index.nlevels}")
    if len(inds) < index.nlevels:
        remainder = [i for i in range(index.nlevels) if i not in inds]
        new_ind: list = [index.map(lambda i: (compressor(i)))] + [index.get_level_values(j) for j in remainder]
        new_names = [compressor(index.names)] + index.names.difference(old_names)
        new_ind: pd.MultiIndex = pd.MultiIndex.from_arrays(new_ind, names=new_names)
    else:
        new_ind: pd.Index = index.map(lambda i: compressor(i))
        new_ind.name = compressor(index.names)

    return new_ind


@_df_loader_wrapper
def rank_by_parameter(basedir: Path, df: pd.DataFrame, rank_on: Any, parameters: list = None, ascending=False,
                      **kwargs) -> pd.DataFrame:
    """ Generate ranks for the Dataframe's rows based on the column specified in 'rank_on'. If a list of additional
    column names is provided, a dataframe containing mean ranks of those parameters is returned, otherwise a copy of
    the initial dataframe with the "ranks" column tacked at the end is returned. """

    ranks = df[rank_on].rank(ascending=ascending, **kwargs).to_frame("rank")
    ranks = df.join([ranks])

    if parameters is not None:
        ranks = ranks.groupby(parameters)["rank"].agg("mean").to_frame("Rank")

    return ranks


@_df_loader_wrapper
def get_nsamples(basedir: Path, df: pd.DataFrame, groupby: list, index: Optional[list] = None, **kwargs) \
        -> pd.DataFrame:
    """
    Returns a DataFrame with a single column - "nsamples" - that contains the number of individual models (as
    identified by the tuple (taskid, model_idx)) that were observed. 'groupby' can be a list of column names that
    dictate the index of the returned DataFrame, i.e. the number of models in each group as identified by a unique
    combination of the respective values in the columns named in 'groupby' is counted. It is assumed that the input
    DataFrame's index does not need to be filtered anymore, e.g. for specific epochs. If it contains epoch-wise data,
    then "nsamples" will count each epoch as a separate data point.

    :param basedir:
    :param df:
    :param groupby: list
        A list of column names, compatible with df.groupby().
    :param index: list of strings
        The new index names - should be a one-to-one renaming of "groupby".
    :param kwargs:
    :return:
    """

    assert all([g in df.columns for g in groupby]), "Mismatch in input DataFrame's columns and given grouping parameters.\nGiven " \
                                  f"parameters:\n{groupby}\n\nDataFrame columns:\n{df.columns}"

    available_cols = df.columns.difference(groupby)
    nsamples = df.xs(1, level=MetricDFIndexLevels.epoch.value)
    nsamples = nsamples[[available_cols[0], *groupby]].groupby(groupby).agg("count")

    if isinstance(nsamples, pd.Series):
        nsamples = nsamples.to_frame("nsamples")
    else:
        assert isinstance(nsamples, pd.DataFrame), f"Expected a pandas DataFrame, but 'nsamples' is {type(nsamples)}"
        assert len(nsamples.columns) == 1, f"Expected a DataFrame with a single column, but the given DataFrame has " \
                                           f"{len(nsamples.columns)} columns."
        nsamples.columns = pd.Index(["nsamples"])

    # Reduce the names of the index levels to the bare minimum necessary
    assert len(index) == len(nsamples.index.names), \
        f"Mismatch in number of index names given - {len(index)} - and the number of available index levels - " \
        f"{nsamples.index.nlevels}"

    nsamples.index = nsamples.index.set_names(index)

    return nsamples


@_df_loader_wrapper
def estimate_remaining_runtime(basedir: Path, df: pd.DataFrame, max_epochs: int = 200) -> pd.DataFrame:
    """
    Given a complete metrics DataFrame, returns a DataFrame containing the estimated runtime needed to finish
    evaluating each model.
    :param df: pandas DataFrame
        A DataFrame object that can be read by the postprocessing modules to extract the number of epochs that each
        model has been evaluated for as well as how long each model has been run for.
    :param max_epochs:
    :return:
    """

    nepochs: pd.DataFrame = get_nepochs(df=df)
    remaining_epochs: pd.DataFrame = nepochs.rsub(max_epochs)

    runtimes: pd.Series = get_runtimes(df=df, reduce_epochs=True)["runtime"]
    runtime_per_epoch: pd.Series = runtimes / nepochs["nepochs"]
    required_runtimes: pd.DataFrame = (remaining_epochs["nepochs"] * runtime_per_epoch).to_frame("required")

    return required_runtimes


@_df_loader_wrapper
def get_op_counts(basedir: Path, df: pd.DataFrame) -> pd.DataFrame:
    """
    Given a complete metrics DataFrame, returns a DataFrame containing the number of times each known operation type
    was used in each model configuration's cell architecture.
    :param df: pandas DataFrame
        A DataFrame object that can be read by the postprocessing modules to extract the number of epochs that each
        model has been evaluated for as well as how long each model has been run for.
    :return:
    """

    configs = get_configs(df=df)
    n_ops = len(EDGE_LIST)
    ops = configs[[f"Op{o+1}" for o in range(n_ops)]]
    max_cols = pd.Int64Index(list(range(len(OP_NAMES))))
    dummies = [pd.get_dummies(ops[c]) for c in ops.columns]
    for d in dummies:
        missing_cols = max_cols.difference(d.columns)
        d[missing_cols] = 0

    freq = sum(dummies)
    freq.columns = freq.columns.map(lambda x: OP_NAMES[x])

    assert (freq.sum(axis=1) == n_ops).all(), "Failed to verify integrity of cell operation frequency calculation."

    return freq


if __name__ == "__main__":
    args = _parse_cli()
    basedir: Path = args.basedir
    get_accuracies(basedir=basedir)

